import json
import os
import mimetypes

from collections import OrderedDict
from enum import Enum
from colorlog import error, warning as warn, info
from urllib.parse import parse_qs
from datetime import datetime
from typing import Dict, List, Optional

from yangson.exceptions import YangsonException, NonexistentSchemaNode, SchemaError, SemanticError
from yangson.schemanode import ContainerNode, ListNode, GroupNode, LeafListNode, LeafNode
from yangson.instance import NonexistentInstance, InstanceValueError, RootNode

from .config import CONFIG_GLOBAL, CONFIG_HTTP, CONFIG_NACM, API_ROOT_data, API_ROOT_RUNNING_data, API_ROOT_ops
from .helpers import CertHelpers, DateTimeHelpers, ErrorHelpers, LogHelpers, SSLCertT
from .handler_list import OP_HANDLERS
from .errors import BackendError
from .nacm import NacmForbiddenError
from .data import (
    BaseDatastore,
    RpcInfo,
    DataLockError,
    NoHandlerError,
    NoHandlerForOpError,
    InstanceAlreadyPresent,
    ChangeType,
    ConfHandlerFailedError,
    OpHandlerFailedError,
    StagingDataException
)

QueryStrT = Dict[str, List[str]]
epretty = ErrorHelpers.epretty
errtag = ErrorHelpers.errtag
debug_httph = LogHelpers.create_module_dbg_logger(__name__)


CT_PLAIN = "text/plain"
CT_YANG_JSON = "application/yang.api+json"


class HttpRequestError(Exception):
    pass


class HttpStatus(Enum):
    Ok          = ("200", "OK")
    Created     = ("201", "Created")
    NoContent   = ("204", "No Content")
    NotModified = ("304", "Not Modified")
    BadRequest  = ("400", "Bad Request")
    Forbidden   = ("403", "Forbidden")
    NotFound    = ("404", "Not Found")
    MethodNotAllowed    = ("405", "Method Not Allowed")
    NotAcceptable       = ("406", "Not Acceptable")
    Conflict    = ("409", "Conflict")
    ReqTooLarge = ("413", "Request Entity Too Large")
    InternalServerError = ("500", "Internal Server Error")

    @property
    def code(self) -> str:
        return self.value[0]

    @property
    def msg(self) -> str:
        return self.value[1]


ERRTAG_MALFORMED = "malformed-message"
ERRTAG_REQLARGE = "request-too-large"
ERRTAG_OPNOTSUPPORTED = "operation-not-supported"
ERRTAG_OPFAILED = "operation-failed"
ERRTAG_ACCDENIED = "access-denied"
ERRTAG_LOCKDENIED = "lock-denied"
ERRTAG_INVVALUE = "invalid-value"
ERRTAG_EXISTS = "data-exists"


class RestconfErrType(Enum):
    Transport   = "transport"
    Rpc         = "rpc"
    Protocol    = "protocol"
    Application = "application"


class HttpResponse:
    def __init__(self, status: HttpStatus, data: bytes, content_type: str, extra_headers: OrderedDict=None):
        self.status_code = status.code
        self.data = data
        self.content_length = len(data)
        self.content_type = content_type
        self.extra_headers = extra_headers

    @classmethod
    def empty(cls, status: HttpStatus, status_in_body: bool=False) -> "HttpResponse":
        if status_in_body:
            response = status.code + " " + status.msg + "\n"
        else:
            response = ""

        return cls(status, response.encode(), CT_PLAIN)

    @classmethod
    def error(cls, status: HttpStatus, err_type: RestconfErrType, err_tag: str, err_apptag: str=None,
              err_path: str=None, err_msg: str=None, exception: Exception=None) -> "HttpResponse":
        err_body = {
            "error-type": err_type.value,
            "error-tag": err_tag
        }

        # Auto-fill app-tag, path and mesage fields from Python's Exception attributes
        if exception is not None:
            try:
                err_body["error-app-tag"] = exception.tag
            except AttributeError:
                pass

            try:
                err_body["error-path"] = exception.path
            except AttributeError:
                pass

            try:
                err_body["error-message"] = exception.__class__.__name__ + ": " + exception.message
            except AttributeError:
                err_body["error-message"] = exception.__class__.__name__ + ": " + str(exception)

        if err_apptag is not None:
            err_body["error-app-tag"] = err_apptag

        if err_path is not None:
            err_body["error-path"] = err_path

        if err_msg is not None:
            err_body["error-message"] = err_msg

        err_template = {
            "ietf-restconf:errors": {
                "error": [
                    err_body
                ]
            }
        }

        response = json.dumps(err_template, indent=4)
        return cls(status, response.encode(), CT_YANG_JSON)


def unknown_req_handler(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT) -> HttpResponse:
    return HttpResponse.error(
        HttpStatus.BadRequest,
        RestconfErrType.Transport,
        ERRTAG_MALFORMED,
        "unknown_req_handler"
    )


def _get_yl_date() -> str:
    try:
        yang_lib_date_ts = os.path.getmtime(os.path.join(CONFIG_GLOBAL["YANG_LIB_DIR"], "yang-library-data.json"))
        yang_lib_date = datetime.fromtimestamp(yang_lib_date_ts).strftime("%Y-%m-%d")
    except OSError:
        yang_lib_date = None

    return yang_lib_date


def api_root_handler(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT):
    # Top level api resource (appendix B.1.1)

    top_res = {
        "ietf-restconf:restconf": {
            "data": {},
            "operations": {},
            "yang-library-version": _get_yl_date()
        }
    }

    response = json.dumps(top_res, indent=4)
    return HttpResponse(HttpStatus.Ok, response.encode(), CT_YANG_JSON)


def api_ylv_handler(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT):
    ylv = {
        "ietf-restconf:yang-library-version": _get_yl_date()
    }

    response = json.dumps(ylv, indent=4)
    return HttpResponse(HttpStatus.Ok, response.encode(), CT_YANG_JSON)


def _get(ds: BaseDatastore, req_headers: OrderedDict, pth: str, username: str, staging: bool=False) -> HttpResponse:
    url_split = pth.split("?")
    url_path = url_split[0]
    if len(url_split) > 1:
        query_string = parse_qs(url_split[1])
    else:
        query_string = {}

    rpc1 = RpcInfo()
    rpc1.username = username
    rpc1.path = url_path.rstrip("/")
    rpc1.qs = query_string

    # Skip NACM check for privileged users
    if username in CONFIG_NACM["ALLOWED_USERS"]:
        rpc1.skip_nacm_check = True

    try:
        ds.lock_data(username)
        http_resp = None

        try:
            n = ds.get_node_rpc(rpc1, staging)
        except NacmForbiddenError as e:
            http_resp = HttpResponse.error(
                HttpStatus.Forbidden,
                RestconfErrType.Protocol,
                ERRTAG_ACCDENIED,
                exception=e
            )
        except (NonexistentSchemaNode, NonexistentInstance) as e:
            http_resp = HttpResponse.error(
                HttpStatus.NotFound,
                RestconfErrType.Protocol,
                ERRTAG_INVVALUE,
                exception=e
            )
        except (InstanceValueError, ValueError) as e:
            http_resp = HttpResponse.error(
                HttpStatus.BadRequest,
                RestconfErrType.Protocol,
                ERRTAG_INVVALUE,
                exception=e
            )
        except (ConfHandlerFailedError, NoHandlerError, YangsonException, BackendError) as e:
            http_resp = HttpResponse.error(
                HttpStatus.InternalServerError,
                RestconfErrType.Protocol,
                ERRTAG_OPFAILED,
                exception=e
            )
        finally:
            ds.unlock_data()

        if http_resp is not None:
            # Return error response
            return http_resp

        hdr_inm = req_headers.get("if-none-match")
        n_etag = str(hash(n.value))

        if (hdr_inm is not None) and (hdr_inm == n_etag):
            http_resp = HttpResponse.empty(HttpStatus.NotModified)
        else:
            n_value = n.raw_value()

            if isinstance(n, RootNode):
                # Getting top-level node
                restconf_env = "ietf-restconf:data"
                restconf_n_value = {restconf_env: n_value}
            else:
                sn = n.schema_node
                if isinstance(sn, (ContainerNode, GroupNode, LeafNode)):
                    restconf_env = "{}:{}".format(sn.qual_name[1], sn.qual_name[0])
                    restconf_n_value = {restconf_env: n_value}
                elif isinstance(sn, ListNode):
                    restconf_env = "{}:{}".format(sn.qual_name[1], sn.qual_name[0])
                    if isinstance(n_value, list):
                        # List and list item points to the same schema node
                        restconf_n_value = {restconf_env: n_value}
                    else:
                        restconf_n_value = {restconf_env: [n_value]}
                else:
                    raise HttpRequestError()

            response = json.dumps(restconf_n_value, indent=4)

            add_headers = OrderedDict()
            add_headers["ETag"] = n_etag
            try:
                lm_time = DateTimeHelpers.to_httpdate_str(n.value.timestamp, CONFIG_GLOBAL["TIMEZONE"])
                add_headers["Last-Modified"] = lm_time
            except AttributeError:
                # Only arrays and objects have last_modified attribute
                pass

            http_resp = HttpResponse(HttpStatus.Ok, response.encode(), CT_YANG_JSON, extra_headers=add_headers)

    except DataLockError as e:
        http_resp = HttpResponse.error(
            HttpStatus.Conflict,
            RestconfErrType.Protocol,
            ERRTAG_LOCKDENIED,
            exception=e
        )
    except HttpRequestError as e:
        http_resp = HttpResponse.error(
            HttpStatus.BadRequest,
            RestconfErrType.Protocol,
            ERRTAG_INVVALUE,
            exception=e
        )

    return http_resp


def create_get_api(ds: BaseDatastore):
    def get_api_closure(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT) -> HttpResponse:
        username = CertHelpers.get_field(client_cert, "emailAddress")
        info("[{}] api_get: {}".format(username, headers[":path"]))

        api_pth = headers[":path"][len(API_ROOT_data):]
        http_resp = _get(ds, headers, api_pth, username, staging=True)
        return http_resp

    return get_api_closure


def create_get_running_api(ds: BaseDatastore):
    def get_staging_api_closure(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT) -> HttpResponse:
        username = CertHelpers.get_field(client_cert, "emailAddress")
        info("[{}] api_get_staging: {}".format(username, headers[":path"]))

        api_pth = headers[":path"][len(API_ROOT_RUNNING_data):]
        http_resp = _get(ds, headers, api_pth, username, staging=False)
        return http_resp

    return get_staging_api_closure


def get_file(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT) -> HttpResponse:
    # Ordinary file on filesystem
    username = CertHelpers.get_field(client_cert, "emailAddress")
    url_path = headers[":path"].split("?")[0]
    url_path_safe = "".join(filter(lambda c: c.isalpha() or c in "/-_.", url_path)).replace("..", "").strip("/")
    file_path = os.path.join(CONFIG_HTTP["DOC_ROOT"], url_path_safe)

    if os.path.isdir(file_path):
        file_path = os.path.join(file_path, CONFIG_HTTP["DOC_DEFAULT_NAME"])

    ctype = mimetypes.guess_type(file_path)[0] or "application/octet-stream"

    try:
        fd = open(file_path, 'rb')
        response = fd.read()
        fd.close()
    except FileNotFoundError:
        warn("[{}] Cannot open requested file \"{}\"".format(username, file_path))
        http_resp = HttpResponse.empty(HttpStatus.NotFound)
    else:
        info("[{}] Serving ordinary file {} of type \"{}\"".format(username, file_path, ctype))
        http_resp = HttpResponse(HttpStatus.Ok, response, ctype)

    return http_resp


def _post(ds: BaseDatastore, pth: str, username: str, data: str) -> HttpResponse:
    debug_httph("HTTP data received: " + data)

    url_split = pth.split("?")
    url_path = url_split[0]
    if len(url_split) > 1:
        query_string = parse_qs(url_split[1])
    else:
        query_string = {}

    rpc1 = RpcInfo()
    rpc1.username = username
    rpc1.path = url_path.rstrip("/")
    rpc1.qs = query_string

    # Skip NACM check for privileged users
    if username in CONFIG_NACM["ALLOWED_USERS"]:
        rpc1.skip_nacm_check = True

    try:
        json_data = json.loads(data) if len(data) > 0 else {}
    except ValueError as e:
        error("Failed to parse POST data: " + epretty(e))
        return HttpResponse.error(
            HttpStatus.BadRequest,
            RestconfErrType.Protocol,
            ERRTAG_INVVALUE,
            exception=e
        )

    try:
        ds.lock_data(username)

        try:
            try:
                staging_root = ds.get_data_root_staging(rpc1.username)
            except StagingDataException:
                info("Starting transaction for user \"{}\"".format(rpc1.username))
                ds.make_user_journal(rpc1.username, None)
                staging_root = ds.get_data_root_staging(rpc1.username)
            new_root = ds.create_node_rpc(staging_root, rpc1, json_data)
            ds.add_to_journal_rpc(ChangeType.CREATE, rpc1, json_data, *new_root)
            http_resp = HttpResponse.empty(HttpStatus.Created)
        except NacmForbiddenError as e:
            http_resp = HttpResponse.error(
                HttpStatus.Forbidden,
                RestconfErrType.Protocol,
                ERRTAG_ACCDENIED,
                exception=e
            )
        except (NonexistentSchemaNode, NonexistentInstance) as e:
            http_resp = HttpResponse.error(
                HttpStatus.NotFound,
                RestconfErrType.Protocol,
                ERRTAG_INVVALUE,
                exception=e
            )
        except NoHandlerError as e:
            http_resp = HttpResponse.error(
                HttpStatus.BadRequest,
                RestconfErrType.Protocol,
                ERRTAG_OPNOTSUPPORTED,
                exception=e
            )
        except (InstanceValueError, YangsonException, ValueError) as e:
            http_resp = HttpResponse.error(
                HttpStatus.BadRequest,
                RestconfErrType.Protocol,
                ERRTAG_INVVALUE,
                exception=e
            )
        except InstanceAlreadyPresent as e:
            http_resp = HttpResponse.error(
                HttpStatus.Conflict,
                RestconfErrType.Protocol,
                ERRTAG_EXISTS,
                exception=e
            )
    except DataLockError as e:
        http_resp = HttpResponse.error(
            HttpStatus.Conflict,
            RestconfErrType.Protocol,
            ERRTAG_LOCKDENIED,
            exception=e
        )
    finally:
        ds.unlock_data()

    return http_resp


def create_post_api(ds: BaseDatastore):
    def post_api_closure(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT) -> HttpResponse:
        username = CertHelpers.get_field(client_cert, "emailAddress")
        info("[{}] api_post: {}".format(username, headers[":path"]))

        api_pth = headers[":path"][len(API_ROOT_data):]
        http_resp = _post(ds, api_pth, username, data)
        return http_resp

    return post_api_closure


def _put(ds: BaseDatastore, pth: str, username: str, data: str) -> HttpResponse:
    debug_httph("HTTP data received: " + data)

    url_split = pth.split("?")
    url_path = url_split[0]

    rpc1 = RpcInfo()
    rpc1.username = username
    rpc1.path = url_path.rstrip("/")

    # Skip NACM check for privileged users
    if username in CONFIG_NACM["ALLOWED_USERS"]:
        rpc1.skip_nacm_check = True

    try:
        json_data = json.loads(data) if len(data) > 0 else {}
    except ValueError as e:
        error("Failed to parse PUT data: " + epretty(e))
        return HttpResponse.error(
            HttpStatus.BadRequest,
            RestconfErrType.Protocol,
            ERRTAG_INVVALUE,
            exception=e
        )

    try:
        ds.lock_data(username)

        try:
            try:
                staging_root = ds.get_data_root_staging(rpc1.username)
            except StagingDataException:
                info("Starting transaction for user \"{}\"".format(rpc1.username))
                ds.make_user_journal(rpc1.username, None)
                staging_root = ds.get_data_root_staging(rpc1.username)
            new_root = ds.update_node_rpc(staging_root, rpc1, json_data)
            ds.add_to_journal_rpc(ChangeType.REPLACE, rpc1, json_data, *new_root)
            http_resp = HttpResponse.empty(HttpStatus.NoContent, status_in_body=False)
        except NacmForbiddenError as e:
            http_resp = HttpResponse.error(
                HttpStatus.Forbidden,
                RestconfErrType.Protocol,
                ERRTAG_ACCDENIED,
                exception=e
            )
        except (NonexistentSchemaNode, NonexistentInstance) as e:
            http_resp = HttpResponse.error(
                HttpStatus.NotFound,
                RestconfErrType.Protocol,
                ERRTAG_INVVALUE,
                exception=e
            )
        except NoHandlerError as e:
            http_resp = HttpResponse.error(
                HttpStatus.BadRequest,
                RestconfErrType.Protocol,
                ERRTAG_OPNOTSUPPORTED,
                exception=e
            )
        except (InstanceValueError, StagingDataException, YangsonException, ValueError) as e:
            http_resp = HttpResponse.error(
                HttpStatus.BadRequest,
                RestconfErrType.Protocol,
                ERRTAG_INVVALUE,
                exception=e
            )
    except DataLockError as e:
        http_resp = HttpResponse.error(
            HttpStatus.Conflict,
            RestconfErrType.Protocol,
            ERRTAG_LOCKDENIED,
            exception=e
        )
    finally:
        ds.unlock_data()

    return http_resp


def create_put_api(ds: BaseDatastore):
    def put_api_closure(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT) -> HttpResponse:
        username = CertHelpers.get_field(client_cert, "emailAddress")
        info("[{}] api_put: {}".format(username, headers[":path"]))

        api_pth = headers[":path"][len(API_ROOT_data):]
        http_resp = _put(ds, api_pth, username, data)
        return http_resp

    return put_api_closure


def _delete(ds: BaseDatastore, pth: str, username: str) -> HttpResponse:
        url_split = pth.split("?")
        url_path = url_split[0]

        rpc1 = RpcInfo()
        rpc1.username = username
        rpc1.path = url_path.rstrip("/")

        # Skip NACM check for privileged users
        if username in CONFIG_NACM["ALLOWED_USERS"]:
            rpc1.skip_nacm_check = True

        try:
            ds.lock_data(username)

            try:
                try:
                    staging_root = ds.get_data_root_staging(rpc1.username)
                except StagingDataException:
                    info("Starting transaction for user \"{}\"".format(rpc1.username))
                    ds.make_user_journal(rpc1.username, None)
                    staging_root = ds.get_data_root_staging(rpc1.username)
                new_root = ds.delete_node_rpc(staging_root, rpc1)
                ds.add_to_journal_rpc(ChangeType.DELETE, rpc1, None, *new_root)
                http_resp = HttpResponse.empty(HttpStatus.NoContent, status_in_body=False)
            except NacmForbiddenError as e:
                http_resp = HttpResponse.error(
                    HttpStatus.Forbidden,
                    RestconfErrType.Protocol,
                    ERRTAG_ACCDENIED,
                    exception=e
                )
            except (NonexistentSchemaNode, NonexistentInstance) as e:
                http_resp = HttpResponse.error(
                    HttpStatus.NotFound,
                    RestconfErrType.Protocol,
                    ERRTAG_INVVALUE,
                    exception=e
                )
            except NoHandlerError as e:
                http_resp = HttpResponse.error(
                    HttpStatus.BadRequest,
                    RestconfErrType.Protocol,
                    ERRTAG_OPNOTSUPPORTED,
                    exception=e
                )
            except (InstanceValueError, StagingDataException, YangsonException) as e:
                http_resp = HttpResponse.error(
                    HttpStatus.BadRequest,
                    RestconfErrType.Protocol,
                    ERRTAG_INVVALUE,
                    exception=e
                )
        except DataLockError as e:
            http_resp = HttpResponse.error(
                HttpStatus.Conflict,
                RestconfErrType.Protocol,
                ERRTAG_LOCKDENIED,
                exception=e
            )
        finally:
            ds.unlock_data()

        return http_resp


def create_api_delete(ds: BaseDatastore):
    def api_delete_closure(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT) -> HttpResponse:
        username = CertHelpers.get_field(client_cert, "emailAddress")
        info("[{}] api_delete: {}".format(username, headers[":path"]))

        api_pth = headers[":path"][len(API_ROOT_data):]
        http_resp = _delete(ds, api_pth, username)
        return http_resp

    return api_delete_closure


def create_api_op(ds: BaseDatastore):
    def api_op_closure(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT) -> HttpResponse:
        username = CertHelpers.get_field(client_cert, "emailAddress")
        info("[{}] invoke_op: {}".format(username, headers[":path"]))

        api_pth = headers[":path"][len(API_ROOT_ops):]
        op_name_fq = api_pth[1:].split("/", maxsplit=1)[0]

        try:
            ns, sel1 = op_name_fq.split(":", maxsplit=1)
        except ValueError:
            return HttpResponse.error(
                HttpStatus.BadRequest,
                RestconfErrType.Protocol,
                ERRTAG_MALFORMED,
                "Operation name must be in fully-qualified format"
            )

        try:
            json_data = json.loads(data) if len(data) > 0 else {}
        except ValueError as e:
            return HttpResponse.error(
                HttpStatus.BadRequest,
                RestconfErrType.Protocol,
                ERRTAG_MALFORMED,
                "Failed to parse POST data: " + epretty(e)
            )

        input_args = json_data.get(ns + ":input")

        rpc1 = RpcInfo()
        rpc1.username = username
        rpc1.path = api_pth
        rpc1.op_name = op_name_fq
        rpc1.op_input_args = input_args

        # Skip NACM check for privileged users
        if username in CONFIG_NACM["ALLOWED_USERS"]:
            rpc1.skip_nacm_check = True

        try:
            ret_data = ds.invoke_op_rpc(rpc1)
            if ret_data is None:
                http_resp = HttpResponse.empty(HttpStatus.NoContent, status_in_body=False)
            else:
                if not isinstance(ret_data, str):
                    response = json.dumps(ret_data, indent=4)
                else:
                    response = ret_data
                http_resp = HttpResponse(HttpStatus.Ok, response.encode(), CT_YANG_JSON)
        except NacmForbiddenError as e:
            http_resp = HttpResponse.error(
                HttpStatus.Forbidden,
                RestconfErrType.Protocol,
                ERRTAG_ACCDENIED,
                exception=e
            )
        except (NonexistentSchemaNode, NonexistentInstance) as e:
            http_resp = HttpResponse.error(
                HttpStatus.NotFound,
                RestconfErrType.Protocol,
                ERRTAG_INVVALUE,
                exception=e
            )
        except InstanceAlreadyPresent as e:
            http_resp = HttpResponse.error(
                HttpStatus.Conflict,
                RestconfErrType.Protocol,
                ERRTAG_EXISTS,
                exception=e
            )
        except NoHandlerForOpError as e:
            http_resp = HttpResponse.error(
                HttpStatus.BadRequest,
                RestconfErrType.Protocol,
                ERRTAG_OPNOTSUPPORTED,
                exception=e
            )
        except (SchemaError, SemanticError) as e:
            http_resp = HttpResponse.error(
                HttpStatus.BadRequest,
                RestconfErrType.Protocol,
                ERRTAG_INVVALUE,
                exception=e
            )
        except (ConfHandlerFailedError, OpHandlerFailedError, StagingDataException, YangsonException) as e:
            http_resp = HttpResponse.error(
                HttpStatus.InternalServerError,
                RestconfErrType.Protocol,
                ERRTAG_OPFAILED,
                exception=e
            )
        except ValueError as e:
            http_resp = HttpResponse.error(
                HttpStatus.BadRequest,
                RestconfErrType.Protocol,
                ERRTAG_INVVALUE,
                exception=e
            )

        return http_resp

    return api_op_closure


def create_api_get_op(ds: BaseDatastore):
    def api_op_closure(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT) -> HttpResponse:
        username = CertHelpers.get_field(client_cert, "emailAddress")
        info("[{}] get_op: {}".format(username, headers[":path"]))

        api_pth = headers[":path"][len(API_ROOT_ops):].rstrip("/")
        op_name_fq = api_pth[1:].split("/", maxsplit=1)[0]

        op_names_dict = dict(map(lambda n: (n[0], None), OP_HANDLERS.handlers))

        if api_pth == "":
            # GET root
            ret_data = {
                "ietf-restconf:operations": op_names_dict
            }
        else:
            # GET particular operation
            try:
                ns, op_name = op_name_fq.split(":", maxsplit=1)
            except ValueError:
                return HttpResponse.error(
                    HttpStatus.BadRequest,
                    RestconfErrType.Protocol,
                    ERRTAG_MALFORMED,
                    "Operation name must be in fully-qualified format"
                )

            if op_names_dict.get(op_name_fq, -1) != -1:
                ret_data = {
                    op_name_fq: None
                }
            else:
                # Not found
                ret_data = None

        if ret_data is None:
            http_resp = HttpResponse.error(
                HttpStatus.NotFound,
                RestconfErrType.Protocol,
                ERRTAG_INVVALUE,
                err_path=api_pth,
                err_msg="Operation name not found"
            )
        else:
            response = json.dumps(ret_data, indent=4)
            http_resp = HttpResponse(HttpStatus.Ok, response.encode(), CT_YANG_JSON)

        return http_resp

    return api_op_closure


def options_api(headers: OrderedDict, data: Optional[str], client_cert: SSLCertT) -> HttpResponse:
    info("api_options: {}".format(headers[":path"]))
    headers_extra = OrderedDict()
    headers_extra["Allow"] = "GET, PUT, POST, OPTIONS, HEAD, DELETE"
    http_resp = HttpResponse(HttpStatus.Ok, bytes(), CT_PLAIN, extra_headers=headers_extra)

    return http_resp




import collections
from io import StringIO
from threading import Lock
from enum import Enum
from typing import Union

from colorlog import error, warning as warn, info
from typing import List, Set, Optional

from yangson.datamodel import DataModel
from yangson.instvalue import Value, ArrayValue, ObjectValue
from yangson.instance import (
    InstanceNode,
    NonexistentSchemaNode,
    NonexistentInstance,
    InstanceRoute,
    MemberName,
    EntryIndex,
    EntryKeys
)

from .helpers import DataHelpers, ErrorHelpers, LogHelpers
from .errors import JetconfError

epretty = ErrorHelpers.epretty
debug_nacm = LogHelpers.create_module_dbg_logger(__name__)


class Action(Enum):
    PERMIT = True
    DENY = False


class Permission(Enum):
    NACM_ACCESS_READ = 0
    NACM_ACCESS_CREATE = 1
    NACM_ACCESS_UPDATE = 2
    NACM_ACCESS_DELETE = 3
    NACM_ACCESS_EXEC = 4


class NacmRuleType(Enum):
    NACM_RULE_NOTSET = 0
    NACM_RULE_OPERATION = 1
    NACM_RULE_NOTIF = 2
    NACM_RULE_DATA = 3


class NacmError(JetconfError):
    pass


class NonexistentUserError(NacmError):
    pass


class NacmForbiddenError(NacmError):
    def __init__(self, msg="Access to data node rejected by NACM", rule=None):
        self.msg = msg
        self.rule = rule

    def __str__(self):
        return "{} (rule: {})".format(self.msg, str(self.rule))


class NacmGroup:
    def __init__(self, name: str, users: List[str]):
        self.name = name
        self.users = users


class NacmRule:
    class TypeData:
        def __init__(self):
            self.path = None        # type: str
            self.rpc_names = None   # type: List[str]
            self.ntf_names = None   # type: List[str]

    def __init__(self):
        self.name = None                            # type: str
        self.comment = None                         # type: str
        self.module = None                          # type: str
        self.type = NacmRuleType.NACM_RULE_NOTSET   # type: NacmRuleType
        self.type_data = self.TypeData()
        self.access = set()                         # type: Set[Permission]
        self.action = Action.DENY


class RuleTreeNode:
    def __init__(self, isel=None, up: "RuleTreeNode"=None):
        self.isel = isel
        self.rule = None    # type: NacmRule
        self.up = up
        self.children = []  # type: List[RuleTreeNode]

    def get_rule(self, perm: Permission) -> Optional[NacmRule]:
        n = self
        while n:
            if (n.rule is not None) and (perm in n.rule.access):
                return n.rule
            n = n.up

        return None

    def get_action(self, perm: Permission) -> Optional[Action]:
        rule = self.get_rule(perm)
        return rule.action if rule is not None else None


class NacmRuleList:
    def __init__(self):
        self.name = ""          # type: str
        self.groups = []        # type: List[NacmGroup]
        self.rules = []         # type: List[NacmRule]


class DataRuleTree:
    def __init__(self, dm: DataModel, rule_lists: List[NacmRuleList]):
        self.root = []  # type: List[RuleTreeNode]

        for rl in rule_lists:
            for rule in filter(lambda r: r.type == NacmRuleType.NACM_RULE_DATA, rl.rules):
                try:
                    ii = dm.parse_instance_id(rule.type_data.path)
                except NonexistentSchemaNode as e:
                    error(epretty(e, __name__))
                    ii = []
                nl = self.root
                node_match_prev = None
                for isel in ii:
                    node_match = (list(filter(lambda x: x.isel == isel, nl)) or [None])[0]
                    if node_match is None:
                        new_elem = RuleTreeNode()
                        new_elem.isel = isel
                        new_elem.up = node_match_prev

                        if isel is ii[-1]:
                            new_elem.rule = rule
                        nl.append(new_elem)
                        node_match_prev = new_elem
                        nl = new_elem.children
                    else:
                        if isel is ii[-1]:
                            node_match.rule = rule
                        node_match_prev = node_match
                        nl = node_match.children

    def _print_rule_tree(self, io_str: StringIO, rule_node_list: List[RuleTreeNode], depth: int, vbars: List[int]):
        indent_str_list = list(("   " * depth) + "+--")
        for vb in vbars:
            indent_str_list[vb * 3] = "|"
        indent_str = "".join(indent_str_list)

        for rule_node in rule_node_list:
            rule = rule_node.rule
            if rule is not None:
                action_str = rule.action.name
                access = sorted(list(map(lambda n: n.name.split("_")[-1].lower(), rule.access)))
                io_str.write(indent_str + " " + str(rule_node.isel) + " " + action_str + str(access) + "\n")
            else:
                io_str.write(indent_str + " " + str(rule_node.isel) + "\n")
            if rule_node is rule_node_list[-1]:
                self._print_rule_tree(io_str, rule_node.children, depth + 1, vbars)
            else:
                self._print_rule_tree(io_str, rule_node.children, depth + 1, vbars + [depth])

    def print_rule_tree(self) -> str:
        io_str = StringIO()
        io_str.write("----- NACM Data Rule tree -----\n")
        self._print_rule_tree(io_str, self.root, 0, [])
        return io_str.getvalue()


class NacmConfig:
    def __init__(self, nacm_ds: "BaseDatastore", dm: DataModel):
        self.nacm_ds = nacm_ds
        self.dm = dm
        self.enabled = False
        self.default_read = Action.PERMIT
        self.default_write = Action.PERMIT
        self.default_exec = Action.PERMIT
        self.nacm_groups = []
        self.rule_lists = []
        self._user_nacm_rpc = {}
        self.internal_data_lock = Lock()
        self._lock_username = None

    # Fills internal read-only data structures
    def update(self):
        lock_res = self.internal_data_lock.acquire(blocking=True, timeout=1)
        if not lock_res:
            error("NACM update: cannot acquire data lock")
            return

        self.nacm_groups = []
        self.rule_lists = []
        self._user_nacm_rpc = {}

        try:
            nacm_json = self.nacm_ds.get_data_root()["ietf-netconf-acm:nacm"].value
        except NonexistentInstance:
            warn("Data does not contain \"ietf-netconf-acm:nacm\" node, NACM rules will be empty")
            return

        self.enabled = nacm_json["enable-nacm"]
        if not self.enabled:
            # NACM not enabled, no need to continue
            self.internal_data_lock.release()
            return

        self.default_read = Action.PERMIT if nacm_json["read-default"] == "permit" else Action.DENY
        self.default_write = Action.PERMIT if nacm_json["write-default"] == "permit" else Action.DENY
        self.default_exec = Action.PERMIT if nacm_json["exec-default"] == "permit" else Action.DENY

        for group in nacm_json["groups"]["group"]:
            self.nacm_groups.append(NacmGroup(group["name"], group["user-name"]))

        for rule_list_json in nacm_json["rule-list"]:
            rl = NacmRuleList()
            rl.name = rule_list_json["name"]
            rl.groups = rule_list_json["group"]

            for rule_json in rule_list_json["rule"]:
                rule = NacmRule()
                rule.name = rule_json.get("name")
                rule.comment = rule_json.get("comment")
                rule.module = rule_json.get("module-name")

                if rule_json.get("access-operations") is not None:
                    access_perm_list = rule_json["access-operations"]
                    if isinstance(access_perm_list, str) and (access_perm_list == "*"):
                        rule.access = set(Permission)
                    elif isinstance(access_perm_list, collections.Iterable):
                        def perm_str2enum(perm_str: str):
                            return {
                                "read": Permission.NACM_ACCESS_READ,
                                "create": Permission.NACM_ACCESS_CREATE,
                                "update": Permission.NACM_ACCESS_UPDATE,
                                "delete": Permission.NACM_ACCESS_DELETE,
                                "exec": Permission.NACM_ACCESS_EXEC,
                            }.get(perm_str)
                        rule.access.update(map(perm_str2enum, access_perm_list))

                if rule_json.get("rpc-name") is not None:
                    if rule.type != NacmRuleType.NACM_RULE_NOTSET:
                        error("Invalid rule definition (multiple cases from rule-type choice): \"{}\"".format(rule.name))
                    else:
                        rule.type = NacmRuleType.NACM_RULE_OPERATION
                        rule.type_data.rpc_names = rule_json.get("rpc-name").split()

                if rule_json.get("notification-name") is not None:
                    if rule.type != NacmRuleType.NACM_RULE_NOTSET:
                        error("Invalid rule definition (multiple cases from rule-type choice): \"{}\"".format(rule.name))
                    else:
                        rule.type = NacmRuleType.NACM_RULE_NOTIF
                        rule.type_data.ntf_names = rule_json.get("notification-name").split()

                if rule_json.get("path") is not None:
                    if rule.type != NacmRuleType.NACM_RULE_NOTSET:
                        error("Invalid rule definition (multiple cases from rule-type choice): \"{}\"".format(rule.name))
                    else:
                        rule.type = NacmRuleType.NACM_RULE_DATA
                        rule.type_data.path = rule_json["path"]

                rule.action = Action.PERMIT if rule_json["action"] == "permit" else Action.DENY
                rl.rules.append(rule)

            self.rule_lists.append(rl)

        self.internal_data_lock.release()

    def create_user_nacm(self, username: str):
        # all_users = set()
        # for gr in self.nacm_groups:
        #     for user in gr.users:
        #         all_users.add(user)

        # for user in all_users:
        #     info("Creating personalized rule list for user \"{}\"".format(user))
        #     self._user_nacm_rpc[user] = UserNacm(self, user)
        # if username not in all_users:
        #     raise NonexistentUserError

        if not self.internal_data_lock.acquire(blocking=True, timeout=1):
            error("Cannot acquire NACM config lock ")
            return

        info("Creating personalized rule list for user \"{}\"".format(username))
        self._user_nacm_rpc[username] = UserRuleSet(self.dm, self, username)

        self.internal_data_lock.release()

    def get_user_rules(self, username: str) -> "UserRuleSet":
        user_nacm = self._user_nacm_rpc.get(username)
        if user_nacm is None:
            self.create_user_nacm(username)
            user_nacm = self._user_nacm_rpc.get(username)

        return user_nacm


# Rules for particular user
class UserRuleSet:
    def __init__(self, dm: DataModel, config: NacmConfig, username: str):
        self.nacm_enabled = config.enabled
        self.default_read = config.default_read
        self.default_write = config.default_write
        self.default_exec = config.default_exec
        self.rule_lists = []

        user_groups = list(filter(lambda x: username in x.users, config.nacm_groups))
        user_groups_names = list(map(lambda x: x.name, user_groups))
        self.rule_lists = list(filter(lambda x: (set(user_groups_names) & set(x.groups)), config.rule_lists))

        self.rule_tree = DataRuleTree(dm, self.rule_lists)
        debug_nacm("Rule tree for user \"{}\":\n{}".format(username, self.rule_tree.print_rule_tree()))

    def check_data_node_permission(self, root: InstanceNode, ii: InstanceRoute, access: Permission) -> Action:
        if not self.nacm_enabled:
            return Action.PERMIT

        data_node_value = (root.value, root.schema_node)

        nl = self.rule_tree.root        # type: List[RuleTreeNode]
        node_match = None               # type: RuleTreeNode
        for isel in ii:
            # Find child by instance selector
            node_match_step = None      # type: RuleTreeNode
            for rule_node in nl:
                if (type(rule_node.isel) == type(isel)) and (rule_node.isel == isel):
                    node_match_step = rule_node
                    break

                if isinstance(isel, EntryIndex) and isinstance(rule_node.isel, EntryKeys) and \
                        (isel.peek_step(*data_node_value)[0] is rule_node.isel.peek_step(*data_node_value)[0]):
                    node_match_step = rule_node
                    break

            if node_match_step:
                nl = node_match_step.children
                node_match = node_match_step
                data_node_value = isel.peek_step(*data_node_value)
            else:
                break

        if node_match is not None:
            # Matching rule found
            retval = node_match.get_action(access)
        else:
            # No matching rule, return default action
            retval = self.default_read if access == Permission.NACM_ACCESS_READ else self.default_write

        # debug_nacm("check_data_node_path, result = {}".format(retval.name))
        return retval

    def _prune_data_tree(self, node: InstanceNode, root: InstanceNode, ii: InstanceRoute, access: Permission) -> InstanceNode:
        if isinstance(node.value, ObjectValue):
            # print("obj: {}".format(node.value))
            nsel = MemberName(name="", ns=None)
            mii = ii + [nsel]
            for child_key in node.value.keys():
                key_splitted = child_key.split(":", maxsplit=1)
                if len(key_splitted) > 1:
                    nsel.namespace, nsel.name = key_splitted
                else:
                    nsel.namespace, nsel.name = (None, key_splitted[0])
                m = nsel.goto_step(node)

                # debug_nacm("checking mii {}".format(mii))
                if self.check_data_node_permission(root, mii, access) == Action.DENY:
                    # debug_nacm("Pruning node {} {}".format(id(node.value[child_key]), node.value[child_key]))
                    debug_nacm("Pruning node {}".format(DataHelpers.ii2str(mii)))
                    node = node.delete_item(child_key)
                else:
                    node = self._prune_data_tree(m, root, mii, access).up()
        elif isinstance(node.value, ArrayValue):
            # print("array: {}".format(node.value))
            nsel = EntryIndex(0)
            eii = ii + [nsel]
            i = 0
            arr_len = len(node.value)
            while i < arr_len:
                nsel.index = i
                e = nsel.goto_step(node)

                # debug_nacm("checking eii {}".format(eii))
                if self.check_data_node_permission(root, eii, access) == Action.DENY:
                    # debug_nacm("Pruning node {} {}".format(id(node.value[i]), node.value[i]))
                    debug_nacm("Pruning node {}".format(DataHelpers.ii2str(eii)))
                    node = node.delete_item(i)
                    arr_len -= 1
                else:
                    i += 1
                    node = self._prune_data_tree(e, root, eii, access).up()

        return node

    def prune_data_tree(self, node: InstanceNode, root: InstanceNode, ii: InstanceRoute, access: Permission) -> InstanceNode:
        if not self.nacm_enabled:
            return node
        else:
            return self._prune_data_tree(node, root, ii, access)

    def check_rpc_name(self, rpc_name: str) -> Action:
        if not self.nacm_enabled:
            return Action.PERMIT

        for rl in self.rule_lists:
            for rpc_rule in filter(lambda r: r.type == NacmRuleType.NACM_RULE_OPERATION, rl.rules):
                if rpc_name in rpc_rule.type_data.rpc_names:
                    return rpc_rule.action

        return self.default_exec



import json
from yangson.datamodel import DataModel

MODULE_DIR = "yang-data"
YANG_LIBRARY = "yang-library-data.json"
DATA_FILE = "data.json"

with open(YANG_LIBRARY) as ylfile:
    yl = ylfile.read()
dm = DataModel(yl, [MODULE_DIR])

with open(DATA_FILE, "rt") as fp:
    json_data = dm.from_raw(json.load(fp))

json_data.validate()
print("end")



import json

from threading import Lock
from enum import Enum
from colorlog import error, warning as warn, info
from typing import List, Any, Dict, Callable, Optional, Tuple
from datetime import datetime

from yangson.datamodel import DataModel
from yangson.enumerations import ContentType, ValidationScope
from yangson.schemanode import (
    SchemaNode,
    ListNode,
    LeafListNode,
    SchemaError,
    SemanticError,
    InternalNode,
    ContainerNode
)
from yangson.instvalue import ArrayValue, ObjectValue
from yangson.instance import (
    InstanceNode,
    NonexistentInstance,
    InstanceValueError,
    MemberName,
    EntryKeys,
    EntryIndex,
    InstanceRoute,
    ArrayEntry,
    RootNode,
    ObjectMember
)

from .helpers import PathFormat, ErrorHelpers, LogHelpers, DataHelpers, JsonNodeT
from .config import CONFIG, CONFIG_NACM
from .nacm import NacmConfig, Permission, Action, NacmForbiddenError
from .handler_list import (
    OP_HANDLERS,
    STATE_DATA_HANDLES,
    CONF_DATA_HANDLES,
    ConfDataObjectHandler,
    ConfDataListHandler,
    StateDataContainerHandler,
    StateDataListHandler
)
from .errors import JetconfError

epretty = ErrorHelpers.epretty
debug_data = LogHelpers.create_module_dbg_logger(__name__)


class ChangeType(Enum):
    CREATE = 0,
    REPLACE = 1,
    DELETE = 2


class DataLockError(JetconfError):
    pass


class StagingDataException(JetconfError):
    pass


class InstanceAlreadyPresent(JetconfError):
    pass


class HandlerError(JetconfError):
    pass


class NoHandlerError(HandlerError):
    pass


class ConfHandlerFailedError(HandlerError):
    pass


class OpHandlerFailedError(HandlerError):
    pass


class NoHandlerForOpError(NoHandlerError):
    def __init__(self, op_name: str):
        self.op_name = op_name

    def __str__(self):
        return "Nonexistent handler for operation \"{}\"".format(self.op_name)


class NoHandlerForStateDataError(NoHandlerError):
    pass


class RpcInfo:
    def __init__(self):
        self.username = None    # type: str
        self.path = None        # type: str
        self.qs = None          # type: Dict[str, List[str]]
        self.path_format = PathFormat.URL   # type: PathFormat
        self.skip_nacm_check = False        # type: bool
        self.op_name = None                 # type: str
        self.op_input_args = None           # type: ObjectValue


class DataChange:
    def __init__(self, change_type: ChangeType, rpc_info: RpcInfo, input_data: JsonNodeT, root_after_change: InstanceNode, nacm_modified: bool):
        self.change_type = change_type
        self.rpc_info = rpc_info
        self.input_data = input_data
        self.root_after_change = root_after_change
        self.nacm_modified = nacm_modified


class UsrChangeJournal:
    def __init__(self, root_origin: InstanceNode, transaction_opts: Optional[JsonNodeT]):
        self._root_origin = root_origin
        self._transaction_opts = transaction_opts
        self._journal = []  # type: List[DataChange]

    def get_root_head(self) -> InstanceNode:
        if len(self._journal) > 0:
            return self._journal[-1].root_after_change
        else:
            return self._root_origin

    def get_root_origin(self) -> InstanceNode:
        return self._root_origin

    def add(self, change: DataChange):
        self._journal.append(change)

    def list(self) -> JsonNodeT:
        changes_info = []
        for ch in self._journal:
            changes_info.append([ch.change_type.name, ch.rpc_info.path])

        return changes_info

    def commit(self, ds: "BaseDatastore") -> bool:
        nacm_modified = False

        if len(self._journal) == 0:
            return False

        if hash(ds.get_data_root()) == hash(self._root_origin):
            info("Commiting new configuration (swapping roots)")
            # Set new root
            nr = self.get_root_head()

            for change in self._journal:
                nacm_modified = nacm_modified or change.nacm_modified
        else:
            info("Commiting new configuration (re-applying changes)")
            nr = ds.get_data_root()

            for change in self._journal:
                nacm_modified = nacm_modified or change.nacm_modified

                if change.change_type == ChangeType.CREATE:
                    nr = ds.create_node_rpc(nr, change.rpc_info, change.input_data)[0]
                elif change.change_type == ChangeType.REPLACE:
                    nr = ds.update_node_rpc(nr, change.rpc_info, change.input_data)[0]
                elif change.change_type == ChangeType.DELETE:
                    nr = ds.delete_node_rpc(nr, change.rpc_info)[0]

        try:
            # Validate syntax and semantics of new data
            if CONFIG["GLOBAL"]["VALIDATE_TRANSACTIONS"] is True:
                nr.validate(ValidationScope.all, ContentType.config)
        except (SchemaError, SemanticError) as e:
            error("Data validation error:")
            error(epretty(e))
            raise e

        # Set new data root
        ds.set_data_root(nr)

        # Update NACM if NACM data has been affected by any edit
        if nacm_modified and ds.nacm is not None:
            ds.nacm.update()

        # Call commit begin hook
        begin_hook_failed = False
        try:
            ds.commit_begin_callback(self._transaction_opts)
        except Exception as e:
            error("Exception occured in commit_begin handler: {}".format(epretty(e)))
            begin_hook_failed = True

        # Run schema node handlers
        conf_handler_failed = False
        if not begin_hook_failed:
            try:
                for change in self._journal:
                    ii = ds.parse_ii(change.rpc_info.path, change.rpc_info.path_format)
                    ds.run_conf_edit_handler(ii, change)
            except Exception as e:
                error("Exception occured in edit handler: {}".format(epretty(e)))
                conf_handler_failed = True

        # Call commit end hook
        end_hook_failed = False
        end_hook_abort_failed = False
        if not (begin_hook_failed or conf_handler_failed):
            try:
                ds.commit_end_callback(self._transaction_opts, failed=False)
            except Exception as e:
                error("Exception occured in commit_end handler: {}".format(epretty(e)))
                end_hook_failed = True

        if begin_hook_failed or conf_handler_failed or end_hook_failed:
            try:
                # Call commit_end callback again with "failed" argument set to True
                ds.commit_end_callback(self._transaction_opts, failed=True)
            except Exception as e:
                error("Exception occured in commit_end handler (abort): {}".format(epretty(e)))
                end_hook_abort_failed = True

        # Return to previous version of data and raise an exception if something went wrong
        if begin_hook_failed or conf_handler_failed or end_hook_failed or end_hook_abort_failed:
            ds.data_root_rollback(history_steps=1, store_current=False)

            # Update NACM again after rollback
            if nacm_modified and ds.nacm is not None:
                ds.nacm.update()

            raise ConfHandlerFailedError("(see logged)")

        return True


class BaseDatastore:
    def __init__(self, dm: DataModel, with_nacm: bool=False):
        def _blankfn(*args, **kwargs):
            pass

        self.name = ""
        self.nacm = None    # type: NacmConfig
        self._data = None   # type: InstanceNode
        self._data_history = []     # type: List[InstanceNode]
        self._yang_lib_data = None  # type: InstanceNode
        self._dm = dm       # type: DataModel
        self._data_lock = Lock()
        self._lock_username = None  # type: str
        self._usr_journals = {}     # type: Dict[str, UsrChangeJournal]
        self.commit_begin_callback = _blankfn   # type: Callable[..., bool]
        self.commit_end_callback = _blankfn     # type: Callable[..., bool]

        if with_nacm:
            self.nacm = NacmConfig(self, self._dm)

        self._yang_lib_data = self._dm.from_raw(self._dm.yang_library)

    # Returns DataModel object
    def get_dm(self) -> DataModel:
        return self._dm

    # Returns the root node of data tree
    def get_data_root(self, previous_version: int=0) -> InstanceNode:
        if previous_version > 0:
            return self._data_history[-previous_version]
        else:
            return self._data

    def get_yl_data_root(self) -> InstanceNode:
        return self._yang_lib_data

    def make_user_journal(self, username: str, transaction_opts: Optional[JsonNodeT]):
        usr_journal = self._usr_journals.get(username)
        if usr_journal is not None:
            raise StagingDataException("Transaction for user \"{}\" already opened".format(username))
        else:
            self._usr_journals[username] = UsrChangeJournal(self._data, transaction_opts)

    def get_user_journal(self, username: str):
        usr_journal = self._usr_journals.get(username)
        if usr_journal is not None:
            return usr_journal
        else:
            raise StagingDataException("Transaction for user \"{}\" not opened".format(username))

    def drop_user_journal(self, username: str):
        usr_journal = self._usr_journals.get(username)
        if usr_journal is not None:
            del self._usr_journals[username]
        else:
            raise StagingDataException("Transaction for user \"{}\" not opened".format(username))

    # Returns the root node of data tree
    def get_data_root_staging(self, username: str) -> InstanceNode:
        usr_journal = self.get_user_journal(username)
        root = usr_journal.get_root_head()
        return root

    # Set a new Instance node as data root, store old root to archive
    def set_data_root(self, new_root: InstanceNode):
        self._data_history.append(self._data)
        self._data = new_root

    def data_root_rollback(self, history_steps: int, store_current: bool):
        if store_current:
            self._data_history.append(self._data)

        self._data = self._data_history[-history_steps]

    def parse_ii(self, path: str, path_format: PathFormat) -> InstanceRoute:
        if path_format == PathFormat.URL:
            ii = self._dm.parse_resource_id(path)
        else:
            ii = self._dm.parse_instance_id(path)

        return ii

    # Get schema node with particular schema address
    def get_schema_node(self, sch_pth: str) -> SchemaNode:
        sn = self._dm.get_data_node(sch_pth)
        if sn is None:
            # raise NonexistentSchemaNode(sch_pth)
            debug_data("Cannot find schema node for " + sch_pth)
        return sn

    # Notify data observers about change in datastore
    def run_conf_edit_handler(self, ii: InstanceRoute, ch: DataChange):
        sch_pth_list = list(filter(lambda n: isinstance(n, MemberName), ii))

        if ch.change_type == ChangeType.CREATE:
            # Get target member name
            input_member_name_fq = tuple(ch.input_data.keys())[0]
            input_member_name_ns, input_member_name = input_member_name_fq.split(":", maxsplit=1)
            # Append it to ii
            sch_pth_list.append(MemberName(input_member_name, None))

        sch_pth = DataHelpers.ii2str(sch_pth_list)
        sn = self.get_schema_node(sch_pth)

        if sn is None:
            return

        h = CONF_DATA_HANDLES.get_handler(str(id(sn)))
        if h is not None:
            info("handler for actual data node triggered")
            if isinstance(h, ConfDataObjectHandler):
                if ch.change_type == ChangeType.CREATE:
                    h.create(ii, ch)
                elif ch.change_type == ChangeType.REPLACE:
                    h.replace(ii, ch)
                elif ch.change_type == ChangeType.DELETE:
                    h.delete(ii, ch)
            if isinstance(h, ConfDataListHandler):
                if ch.change_type == ChangeType.CREATE:
                    h.create_item(ii, ch)
                elif ch.change_type == ChangeType.REPLACE:
                    h.replace_item(ii, ch)
                elif ch.change_type == ChangeType.DELETE:
                    h.delete_item(ii, ch)
        else:
            sn = sn.parent
            while sn is not None:
                h = CONF_DATA_HANDLES.get_handler(str(id(sn)))
                if h is not None and isinstance(h, ConfDataObjectHandler):
                    info("handler for superior data node triggered, replace")
                    # print(h.schema_path)
                    # print(h.__class__.__name__)
                    h.replace(ii, ch)
                if h is not None and isinstance(h, ConfDataListHandler):
                    info("handler for superior data node triggered, replace_item")
                    h.replace_item(ii, ch)
                sn = sn.parent

    # Get data node, evaluate NACM if required
    def get_node_rpc(self, rpc: RpcInfo, staging=False) -> InstanceNode:
        ii = self.parse_ii(rpc.path, rpc.path_format)

        if staging:
            try:
                root = self.get_data_root_staging(rpc.username)
            except StagingDataException:
                # root = self._data
                info("Starting transaction for user \"{}\"".format(rpc.username))
                self.make_user_journal(rpc.username, None)
                root = self.get_data_root_staging(rpc.username)
        else:
            root = self._data

        yl_data_request = False
        if (len(ii) > 0) and (isinstance(ii[0], MemberName)):
            # Not getting root
            ns_first = ii[0].namespace
            if (ns_first == "ietf-netconf-acm") and (rpc.username not in CONFIG_NACM["ALLOWED_USERS"]):
                raise NacmForbiddenError(rpc.username + " not allowed to access NACM data")
            elif ns_first == "ietf-yang-library":
                root = self._yang_lib_data
                yl_data_request = True
        else:
            # Root node requested
            # Remove NACM data if user is not NACM privieged
            if rpc.username not in CONFIG_NACM["ALLOWED_USERS"]:
                try:
                    root = root.delete_item("ietf-netconf-acm:nacm")
                except NonexistentInstance:
                    pass

            # Append YANG library data
            for member_name, member_val in self._yang_lib_data.value.items():
                root = root.put_member(member_name, member_val).top()

        # Resolve schema node of the desired data node
        sch_pth_list = filter(lambda isel: isinstance(isel, MemberName), ii)
        sch_pth = DataHelpers.ii2str(sch_pth_list)
        sn = self.get_schema_node(sch_pth)

        state_roots = sn.state_roots()

        # Check if URL points to state data or node that contains state data
        if state_roots and not yl_data_request:
            debug_data("State roots: {}".format(state_roots))
            n = None

            for state_root_sch_pth in state_roots:
                state_root_sn = self._dm.get_data_node(state_root_sch_pth)

                # Check if the desired node is child of the state root
                sni = sn
                is_child = False
                while sni:
                    if sni is state_root_sn:
                        is_child = True
                        break
                    sni = sni.parent

                if is_child:
                    # Direct request for the state data
                    sdh = STATE_DATA_HANDLES.get_handler(state_root_sch_pth)
                    if sdh is not None:
                        if isinstance(sdh, StateDataContainerHandler):
                            state_handler_val = sdh.generate_node(ii, rpc.username, staging)
                            state_root_n = sdh.schema_node.orphan_instance(state_handler_val)
                        elif isinstance(sdh, StateDataListHandler):
                            if (sn is sdh.schema_node) and isinstance(ii[-1], MemberName):
                                state_handler_val = sdh.generate_list(ii, rpc.username, staging)
                                state_root_n = sdh.schema_node.orphan_instance(state_handler_val)
                            else:
                                state_handler_val = sdh.generate_item(ii, rpc.username, staging)
                                state_root_n = sdh.schema_node.orphan_entry(state_handler_val)

                        # Select desired subnode from handler-generated content
                        ii_prefix, ii_rel = sdh.schema_node.split_instance_route(ii)
                        n = state_root_n.goto(ii_rel)

                        # There should be only one state root, no need to continue
                        if len(state_roots) != 1:
                            warn("URI points to directly to state data, but more state roots found")
                        break
                    else:
                        raise NoHandlerForStateDataError(rpc.path)
                else:
                    # Request for config data containing state data
                    n = root.goto(ii)

                    def _fill_state_roots(node: InstanceNode) -> InstanceNode:
                        if isinstance(node.value, ObjectValue):
                            if node.schema_node is state_root_sn.parent:
                                ii_gen = DataHelpers.node_get_ii(node)
                                sdh = STATE_DATA_HANDLES.get_handler(state_root_sch_pth)
                                if sdh is not None:
                                    try:
                                        if isinstance(sdh, StateDataContainerHandler):
                                            state_handler_val = sdh.generate_node(ii_gen, rpc.username, staging)
                                        elif isinstance(sdh, StateDataListHandler):
                                            state_handler_val = sdh.generate_list(ii_gen, rpc.username, staging)
                                    except Exception as e:
                                        error("Error occured in state data generator (sn: {})".format(state_root_sch_pth))
                                        error(epretty(e))
                                        error("This state node will be omitted.")
                                    else:
                                        if state_root_sn.ns == state_root_sn.parent.ns:
                                            nm_name = state_root_sn.qual_name[0]
                                        else:
                                            nm_name = state_root_sn.qual_name[1] + ":" + state_root_sn.qual_name[0]

                                        # print("nm={}".format(nm_name))
                                        node = node.put_member(nm_name, state_handler_val, raw=True).up()
                            else:
                                for key in node:
                                    member = node[key]
                                    node = _fill_state_roots(member).up()
                        elif isinstance(node.value, ArrayValue):
                            i = 0
                            arr_len = len(node.value)
                            while i < arr_len:
                                node = _fill_state_roots(node[i]).up()
                                i += 1

                        return node

                    n = _fill_state_roots(n)
                    root = n.top()
        else:
            # No state data in requested node
            n = root.goto(ii)

        # Process "with-defaults" query parameter
        try:
            with_defs = rpc.qs["with-defaults"][0]
        except (IndexError, KeyError):
            with_defs = None

        if with_defs == "report-all":
            n = n.add_defaults()

        # Evaluate NACM if required
        if self.nacm and not rpc.skip_nacm_check:
            nrpc = self.nacm.get_user_rules(rpc.username)
            if nrpc.check_data_node_permission(root, ii, Permission.NACM_ACCESS_READ) == Action.DENY:
                raise NacmForbiddenError()
            else:
                # Prune nodes that should not be accessible to user
                n = nrpc.prune_data_tree(n, root, ii, Permission.NACM_ACCESS_READ)

        # Process "depth" query parameter
        try:
            max_depth_str = rpc.qs["depth"][0]
            if max_depth_str == "unbounded":
                max_depth = None
            else:
                max_depth = int(max_depth_str) - 1
                if (max_depth < 0) or (max_depth > 65535):
                    raise ValueError()
        except (IndexError, KeyError):
            max_depth = None
        except ValueError:
            raise ValueError("Invalid value of query param \"depth\"")

        if max_depth is not None:
            def _tree_limit_depth(node: InstanceNode, depth: int) -> InstanceNode:
                if isinstance(node.value, ObjectValue):
                    if depth > max_depth:
                        node.value = ObjectValue({})
                    else:
                        for child_key in sorted(node.value.keys()):
                            m = node[child_key]
                            node = _tree_limit_depth(m, depth + 1).up()
                elif isinstance(node.value, ArrayValue):
                    depth -= 1
                    for i in range(len(node.value)):
                        e = node[i]
                        node = _tree_limit_depth(e, depth + 1).up()

                return node
            n = _tree_limit_depth(n, 1)

        # Return result
        return n

    # Create new data node
    def create_node_rpc(self, root: InstanceNode, rpc: RpcInfo, value: Any) -> Tuple[InstanceNode, bool]:
        ii = self.parse_ii(rpc.path, rpc.path_format)

        # Get target member name
        input_member_keys = tuple(value.keys())
        if len(input_member_keys) != 1:
            raise ValueError("Received json object must contain exactly one member")

        input_member_name_fq = input_member_keys[0]
        try:
            input_member_ns, input_member_name = input_member_name_fq.split(":", maxsplit=1)
        except ValueError:
            raise ValueError("Input object name must me in fully-qualified format")
        input_member_value = value[input_member_name_fq]

        # Deny any changes of NACM data for non-privileged users
        nacm_changed = False
        if (len(ii) > 0) and (isinstance(ii[0], MemberName)):
            # Not getting root
            ns_first = ii[0].namespace
            if ns_first == "ietf-netconf-acm":
                nacm_changed = True
                if rpc.username not in CONFIG_NACM["ALLOWED_USERS"]:
                    raise NacmForbiddenError(rpc.username + " not allowed to modify NACM data")
        else:
            # Editing root node
            if input_member_ns == "ietf-netconf-acm":
                nacm_changed = True
                if rpc.username not in CONFIG_NACM["ALLOWED_USERS"]:
                    raise NacmForbiddenError(rpc.username + " not allowed to modify NACM data")

        # Evaluate NACM
        if self.nacm and not rpc.skip_nacm_check:
            nrpc = self.nacm.get_user_rules(rpc.username)
            if nrpc.check_data_node_permission(root, ii, Permission.NACM_ACCESS_CREATE) == Action.DENY:
                raise NacmForbiddenError()

        n = root.goto(ii)

        # Get target schema node
        sn = n.schema_node  # type: InternalNode
        member_sn = sn.get_child(input_member_name, input_member_ns)

        if member_sn is None:
            raise ValueError("Received json object contains unknown member")

        # Check if target member already exists
        if sn.ns == member_sn.ns:
            try:
                existing_member = n[input_member_name]
            except NonexistentInstance:
                existing_member = None
        else:
            try:
                existing_member = n[input_member_name_fq]
            except NonexistentInstance:
                existing_member = None

        # Get query parameters
        insert = rpc.qs.get("insert", [None])[0]
        point = rpc.qs.get("point", [None])[0]

        if isinstance(member_sn, ListNode):
            # Append received node to list

            # Create list if necessary
            if existing_member is None:
                new_member_name = input_member_name if n.namespace == input_member_ns else input_member_name_fq
                existing_member = n.put_member(new_member_name, ArrayValue([]))

            # Get ListNode key names
            list_node_keys = member_sn.keys     # Key names in the form [(key, ns), ]

            if insert == "first":
                # Optimization
                if len(existing_member.value) > 0:
                    list_entry_first = existing_member[0]   # type: ArrayEntry
                    new_member = list_entry_first.insert_before(input_member_value, raw=True).up()
                else:
                    new_member = existing_member.update([input_member_value], raw=True)
            elif (insert == "last") or (insert is None):
                # Optimization
                if len(existing_member.value) > 0:
                    list_entry_last = existing_member[-1]   # type: ArrayEntry
                    new_member = list_entry_last.insert_after(input_member_value, raw=True).up()
                else:
                    new_member = existing_member.update([input_member_value], raw=True)
            elif (insert == "before") and (point is not None):
                point_keys_val = point.split(",")  # List key values passed in the "point" query argument
                if len(list_node_keys) != len(point_keys_val):
                    raise ValueError(
                        "Invalid number of keys passed in 'point' query: {} ({} expected)".format(
                            len(point_keys_val), len(list_node_keys)
                        )
                    )
                entry_keys = dict(map(lambda i: (list_node_keys[i], point_keys_val[i]), range(len(list_node_keys))))
                entry_sel = EntryKeys(entry_keys)
                point_list_entry = entry_sel.goto_step(existing_member)   # type: ArrayEntry
                new_member = point_list_entry.insert_before(input_member_value, raw=True).up()
            elif (insert == "after") and (point is not None):
                point_keys_val = point.split(",")  # List key values passed in the "point" query argument
                if len(list_node_keys) != len(point_keys_val):
                    raise ValueError(
                        "Invalid number of keys passed in 'point' query: {} ({} expected)".format(
                            len(point_keys_val), len(list_node_keys)
                        )
                    )
                entry_keys = dict(map(lambda i: (list_node_keys[i], point_keys_val[i]), range(len(list_node_keys))))
                entry_sel = EntryKeys(entry_keys)
                point_list_entry = entry_sel.goto_step(existing_member)   # type: ArrayEntry
                new_member = point_list_entry.insert_after(input_member_value, raw=True).up()
            else:
                raise ValueError("Invalid 'insert'/'point' query values")
        elif isinstance(member_sn, LeafListNode):
            # Append received node to leaf list

            # Create leaf list if necessary
            if existing_member is None:
                new_member_name = input_member_name if n.namespace == input_member_ns else input_member_name_fq
                existing_member = n.put_member(new_member_name, ArrayValue([]))

            # Convert input data from List/Dict to ArrayValue/ObjectValue
            new_value_item = member_sn.entry_from_raw(input_member_value)

            if insert == "first":
                new_member = existing_member.update(ArrayValue([new_value_item] + existing_member.value))
            elif (insert == "last") or (insert is None):
                new_member = existing_member.update(ArrayValue(existing_member.value + [new_value_item]))
            else:
                raise ValueError("Invalid 'insert' query value")
        else:
            # Create new container member

            if existing_member is None:
                # Create new node (object member)
                new_member_name = input_member_name if n.namespace == input_member_ns else input_member_name_fq
                new_member = n.put_member(new_member_name, input_member_value, raw=True)
            else:
                # Data node already exists
                raise InstanceAlreadyPresent("Member \"{}\" already present in \"{}\"".format(input_member_name, ii))

        return new_member.top(), nacm_changed

    # PUT data node
    def update_node_rpc(self, root: InstanceNode, rpc: RpcInfo, value: Any) -> Tuple[InstanceNode, bool]:
        ii = self.parse_ii(rpc.path, rpc.path_format)

        # Get target member name
        input_member_keys = tuple(value.keys())
        if len(input_member_keys) != 1:
            raise ValueError("Received json object must contain exactly one member")

        input_member_name_fq = input_member_keys[0]
        try:
            input_member_ns, input_member_name = input_member_name_fq.split(":", maxsplit=1)
        except ValueError:
            raise ValueError("Input object name must me in fully-qualified format")
        input_member_value = value[input_member_name_fq]

        n = root.goto(ii)

        # Deny any changes of NACM data for non-privileged users
        nacm_changed = False
        if (len(ii) > 0) and (isinstance(ii[0], MemberName)):
            # Not getting root
            ns_first = ii[0].namespace
            if ns_first == "ietf-netconf-acm":
                nacm_changed = True
                if rpc.username not in CONFIG_NACM["ALLOWED_USERS"]:
                    raise NacmForbiddenError(rpc.username + " not allowed to modify NACM data")
        else:
            # Replacing root node
            # Check if NACM data are present in the datastore
            nacm_val = n.value.get("ietf-netconf-acm:nacm")
            if nacm_val is not None:
                nacm_changed = True
                if rpc.username not in CONFIG_NACM["ALLOWED_USERS"]:
                    raise NacmForbiddenError(rpc.username + " not allowed to modify NACM data")

        # Evaluate NACM
        if self.nacm and not rpc.skip_nacm_check:
            nrpc = self.nacm.get_user_rules(rpc.username)
            if nrpc.check_data_node_permission(root, ii, Permission.NACM_ACCESS_UPDATE) == Action.DENY:
                raise NacmForbiddenError()

        new_n = n.update(input_member_value, raw=True)
        new_n.validate(ValidationScope.syntax)

        return new_n.top(), nacm_changed

    # Delete data node
    def delete_node_rpc(self, root: InstanceNode, rpc: RpcInfo) -> Tuple[InstanceNode, bool]:
        ii = self.parse_ii(rpc.path, rpc.path_format)
        n = root.goto(ii)

        # Deny any changes of NACM data for non-privileged users
        nacm_changed = False
        if (len(ii) > 0) and (isinstance(ii[0], MemberName)):
            # Not getting root
            ns_first = ii[0].namespace
            if ns_first == "ietf-netconf-acm":
                nacm_changed = True
                if rpc.username not in CONFIG_NACM["ALLOWED_USERS"]:
                    raise NacmForbiddenError(rpc.username + " not allowed to modify NACM data")
        else:
            # Deleting root node
            # Check if NACM data are present in the datastore
            nacm_val = n.value.get("ietf-netconf-acm:nacm")
            if nacm_val is not None:
                nacm_changed = True
                if rpc.username not in CONFIG_NACM["ALLOWED_USERS"]:
                    raise NacmForbiddenError(rpc.username + " not allowed to modify NACM data")

        # Evaluate NACM
        if self.nacm and not rpc.skip_nacm_check:
            nrpc = self.nacm.get_user_rules(rpc.username)
            if nrpc.check_data_node_permission(root, ii, Permission.NACM_ACCESS_DELETE) == Action.DENY:
                raise NacmForbiddenError()

        if len(ii) == 0:
            # Deleting entire datastore
            new_n = RootNode(ObjectValue({}), root.schema_node, datetime.now())
        else:
            n_parent = n.up()
            last_isel = ii[-1]
            if isinstance(n_parent.value, ArrayValue):
                if isinstance(last_isel, EntryIndex):
                    new_n = n_parent.delete_item(last_isel.index)
                elif isinstance(last_isel, EntryKeys):
                    new_n = n_parent.delete_item(n.index)
                else:
                    raise ValueError("Unknown node selector")
            elif isinstance(n_parent.value, ObjectValue):
                new_n = n_parent.delete_item(last_isel.namespace + ":" + last_isel.name if last_isel.namespace else last_isel.name)
            else:
                raise InstanceValueError(rpc.path, "Invalid target node type")

        return new_n.top(), nacm_changed

    # Invoke an operation
    def invoke_op_rpc(self, rpc: RpcInfo) -> JsonNodeT:
        if rpc.op_name.startswith("jetconf:"):
            # Jetconf internal operation
            op_handler = OP_HANDLERS.get_handler(rpc.op_name)
            if op_handler is None:
                raise NoHandlerForOpError(rpc.op_name)

            ret_data = op_handler(rpc)
        else:
            # External operation defined in data model
            if self.nacm and not rpc.skip_nacm_check:
                nrpc = self.nacm.get_user_rules(rpc.username)
                if nrpc.check_rpc_name(rpc.op_name) == Action.DENY:
                    raise NacmForbiddenError(
                        "Invocation of \"{}\" operation denied for user \"{}\"".format(rpc.op_name, rpc.username)
                    )

            op_handler = OP_HANDLERS.get_handler(rpc.op_name)
            if op_handler is None:
                raise NoHandlerForOpError(rpc.op_name)

            # Get operation input schema
            sn = self._dm.get_schema_node(rpc.path)
            sn_input = sn.get_child("input")

            # Input arguments are expected, this will validate them
            op_input_args = sn_input.from_raw(rpc.op_input_args) if sn_input.children else None

            try:
                ret_data = op_handler(op_input_args, rpc.username)
            except Exception as e:
                raise OpHandlerFailedError(epretty(e))

        return ret_data

    def add_to_journal_rpc(self, ch_type: ChangeType, rpc: RpcInfo, value: Optional[JsonNodeT], new_root: InstanceNode, nacm_modified: bool):
        usr_journal = self._usr_journals.get(rpc.username)
        if usr_journal is not None:
            usr_journal.add(DataChange(ch_type, rpc, value, new_root, nacm_modified))
        else:
            raise NoHandlerError("No active changelist for user \"{}\"".format(rpc.username))

    # Locks datastore data
    def lock_data(self, username: str = None, blocking: bool=True):
        ret = self._data_lock.acquire(blocking=blocking, timeout=1)
        if ret:
            self._lock_username = username or "(unknown)"
            debug_data("Acquired lock in datastore \"{}\" for user \"{}\"".format(self.name, username))
        else:
            raise DataLockError(
                "Failed to acquire lock in datastore \"{}\" for user \"{}\", already locked by \"{}\"".format(
                    self.name,
                    username,
                    self._lock_username
                )
            )

    # Unlock datastore data
    def unlock_data(self):
        self._data_lock.release()
        debug_data("Released lock in datastore \"{}\" for user \"{}\"".format(self.name, self._lock_username))
        self._lock_username = None

    # Load data from persistent storage
    def load(self):
        raise NotImplementedError("Not implemented in base class")

    # Save data to persistent storage
    def save(self):
        raise NotImplementedError("Not implemented in base class")


class JsonDatastore(BaseDatastore):
    def __init__(self, dm: DataModel, json_file: str, with_nacm: bool=False):
        super().__init__(dm, with_nacm)
        self.json_file = json_file

    def load(self):
        self._data = None
        with open(self.json_file, "rt") as fp:
            self._data = self._dm.from_raw(json.load(fp))

        if self.nacm is not None:
            self.nacm.update()

    def save(self):
        with open(self.json_file, "w") as jfd:
            json.dump(self._data.raw_value(), jfd, indent=4)



from yangson.instance import InstanceRoute, NonexistentInstance


# Base class for all exceptions defined in jetconf
class JetconfError(Exception):
    def __init__(self, msg=""):
        self.msg = msg

    def __str__(self):
        return self.msg


class BackendError(JetconfError):
    pass


class StateNonexistentInstance(NonexistentInstance):
    def __init__(self, ii: InstanceRoute, text: str) -> None:
        self.ii = ii
        self.text = text

    def __str__(self):
        return str(self.ii) + ": " + self.text



import os
import yaml

from colorlog import info

_yang_mod_dir_env = os.environ.get("YANG_MODPATH")

CONFIG_GLOBAL = {
    "TIMEZONE": "GMT",
    "LOGFILE": "-",
    "PIDFILE": "/tmp/jetconf.pid",
    "PERSISTENT_CHANGES": True,
    "LOG_LEVEL": "info",
    "LOG_DBG_MODULES": ["*"],
    "YANG_LIB_DIR": _yang_mod_dir_env,
    "DATA_JSON_FILE": "data.json",
    "VALIDATE_TRANSACTIONS": True,
    "BACKEND_PACKAGE": "jetconf_jukebox"
}

CONFIG_HTTP = {
    "DOC_ROOT": "doc-root",
    "DOC_DEFAULT_NAME": "index.html",
    "API_ROOT": "/restconf",
    "API_ROOT_RUNNING": "/restconf_running",
    "SERVER_NAME": "jetconf-h2",
    "UPLOAD_SIZE_LIMIT": 1,
    "LISTEN_LOCALHOST_ONLY": False,
    "PORT": 8443,

    "SERVER_SSL_CERT": "server.crt",
    "SERVER_SSL_PRIVKEY": "server.key",
    "CA_CERT": "ca.pem",
    "DBG_DISABLE_CERTS": False
}

CONFIG_NACM = {
    "ENABLED": True,
    "ALLOWED_USERS": []
}

CONFIG_KNOT = {
    "SOCKET": "/tmp/knot.sock"
}

CONFIG = {
    "GLOBAL": CONFIG_GLOBAL,
    "HTTP_SERVER": CONFIG_HTTP,
    "NACM": CONFIG_NACM,
    "KNOT": CONFIG_KNOT
}

API_ROOT_data = os.path.join(CONFIG_HTTP["API_ROOT"], "data")
API_ROOT_RUNNING_data = os.path.join(CONFIG_HTTP["API_ROOT_RUNNING"], "data")
API_ROOT_ops = os.path.join(CONFIG_HTTP["API_ROOT"], "operations")
API_ROOT_ylv = os.path.join(CONFIG_HTTP["API_ROOT"], "yang-library-version")


def load_config(filename: str) -> bool:
    global API_ROOT_data
    global API_ROOT_RUNNING_data
    global API_ROOT_ops
    global API_ROOT_ylv

    with open(filename) as conf_fd:
        conf_yaml = yaml.load(conf_fd)
        for conf_key in CONFIG.keys():
            try:
                CONFIG[conf_key].update(conf_yaml[conf_key])
            except KeyError:
                pass

    # Shortcuts
    API_ROOT_data = os.path.join(CONFIG_HTTP["API_ROOT"], "data")
    API_ROOT_RUNNING_data = os.path.join(CONFIG_HTTP["API_ROOT_RUNNING"], "data")
    API_ROOT_ops = os.path.join(CONFIG_HTTP["API_ROOT"], "operations")
    API_ROOT_ylv = os.path.join(CONFIG_HTTP["API_ROOT"], "yang-library-version")


def validate_config():
    if CONFIG_GLOBAL["YANG_LIB_DIR"] is None:
        raise ValueError("YANG module directory must be specified (in config file or YANG_MODPATH env variable)")


def print_config():
    info("Using config:\n" + yaml.dump(CONFIG, default_flow_style=False))



import logging

from colorlog import debug, getLogger
from enum import Enum
from typing import List, Dict, Union, Any, Iterable
from datetime import datetime
from pytz import timezone
from yangson.instance import InstanceRoute, MemberName, EntryKeys, InstanceNode, ArrayValue, NonexistentInstance
from yangson.schemanode import ListNode, ContainerNode

from .config import CONFIG_GLOBAL, CONFIG_HTTP

JsonNodeT = Union[Dict[str, Any], List]
SSLCertT = Dict[str, Any]


class PathFormat(Enum):
    URL = 0
    XPATH = 1


class CertHelpers:
    @staticmethod
    def get_field(cert: SSLCertT, key: str) -> str:
        if CONFIG_HTTP["DBG_DISABLE_CERTS"] and (key == "emailAddress"):
            return "test-user"

        try:
            retval = ([x[0][1] for x in cert["subject"] if x[0][0] == key] or [None])[0]
        except (IndexError, KeyError, TypeError):
            retval = None
        return retval


class DataHelpers:
    # Create parent data nodes to JSON subtree up to top level
    @staticmethod
    def node2doc(ii: InstanceRoute, val: Any) -> Dict[str, Any]:
        n = val
        for isel in reversed(ii):
            if isinstance(isel, MemberName):
                n = {isel.key: n}
            if isinstance(isel, EntryKeys):
                n.update(isel.keys)
                n = [n]
        return n

    # Get the namespace of the first segment in path
    # Raises ValueError if the first segment is not in fully-qualified format
    # Returns empty string if api_pth is empty
    @staticmethod
    def path_first_ns(api_pth: str) -> str:
        if (len(api_pth) > 0) and (api_pth[0] == "/"):
            first_seg = api_pth[1:].split("/", maxsplit=1)[0]
            ns1, sel1 = first_seg.split(":", maxsplit=1)
        else:
            ns1 = ""
        return ns1

    # Convert InstanceRoute or List[InstanceSelector] to string
    @staticmethod
    def ii2str(ii: Iterable) -> str:
        return "".join([str(seg) for seg in ii])

    @staticmethod
    def node_get_ii(node: InstanceNode) -> InstanceRoute:
        m = node
        ii_gen = InstanceRoute()
        try:
            while m:
                m_sn = m.schema_node
                m_sn_dp = m_sn.data_parent()

                if isinstance(m_sn, ListNode):
                    if isinstance(m.value, ArrayValue):
                        mn = MemberName(m_sn.qual_name[0], None if m_sn_dp and m_sn.ns == m_sn_dp.ns else m_sn.qual_name[1])
                        ii_gen.append(mn)
                    else:
                        kv = {}
                        for qk in m_sn.keys:
                            k = qk[0]
                            k_ns = None if m_sn_dp and m_sn.ns == m_sn_dp.ns else qk[1]
                            kv[(k, k_ns)] = m.value.get(k)
                        ek = EntryKeys(kv)
                        ii_gen.append(ek)
                elif isinstance(m_sn, ContainerNode):
                    mn = MemberName(m_sn.qual_name[0], None if m_sn_dp and m_sn.ns == m_sn_dp.ns else m_sn.qual_name[1])
                    ii_gen.append(mn)
                m = m.up()
        except NonexistentInstance:
            pass

        ii_gen.reverse()
        return ii_gen


class DateTimeHelpers:
    @staticmethod
    def to_httpdate_str(dt: datetime, local_tz: str=None) -> str:
        if local_tz is not None:
            dtl = timezone(local_tz).localize(dt)
            dt_gmt = dtl.astimezone(timezone("GMT"))
        else:
            dt_gmt = dt

        return dt_gmt.strftime("%a, %d %b %Y %H:%M:%S GMT")


class ErrorHelpers:
    @staticmethod
    def epretty(e: BaseException, module_name: str=None) -> str:
        err_str = e.__class__.__name__ + ": " + str(e)
        if module_name is not None:
            return "In module " + module_name + ": " + err_str
        else:
            return err_str

    @staticmethod
    def errtag(e: BaseException) -> str:
        try:
            tag = e.tag
        except AttributeError:
            tag = None

        return tag


class LogHelpers:
    @staticmethod
    def create_module_dbg_logger(module_name: str):
        module_name_simple = module_name.split(".")[-1]

        def module_dbg_logger(msg: str):
            if ({module_name_simple, "*"} & set(CONFIG_GLOBAL["LOG_DBG_MODULES"])) and (CONFIG_GLOBAL["LOG_LEVEL"] == "debug"):
                logger = getLogger()
                logger.setLevel(logging.DEBUG)
                debug(module_name_simple + ": " + msg)
                logger.setLevel(logging.INFO)

        return module_dbg_logger



import json
import pytest
from jetconf.helpers import DataHelpers
from jetconf.data import JsonDatastore, RpcInfo, PathFormat
from jetconf.nacm import NacmConfig, Permission, Action


@pytest.fixture
def data_model():
    return DataHelpers.load_data_model("./data", "./data/yang-library-data.json")


@pytest.fixture
def datastore_1(data_model):
    ds = JsonDatastore(data_model)
    ds.load("jetconf/example-data.json")
    return ds


@pytest.fixture
def nacm_datastore_1(data_model):
    ds = JsonDatastore(data_model)
    ds.load("jetconf/example-data-nacm.json")
    return ds


def test_datastore(datastore_1):
    data = datastore_1
    rpc = RpcInfo()
    rpc.username = "dominik"
    rpc.path = "/dns-server:dns-server/zones/zone[domain='example.com']/query-module"
    rpc.path_format = PathFormat.XPATH

    # info("Testing read of " + rpc.path)
    n = data.get_node_rpc(rpc)

    expected_value = \
        [
            {'name': 'test1', 'type': 'knot-dns:synth-record'},
            {'name': 'test2', 'type': 'knot-dns:synth-record'}
        ]

    assert json.loads(json.dumps(n.value)) == expected_value

    rpc.path = "/dns-server:dns-server/zones"
    rpc.path_format = PathFormat.URL

    # info("Testing creation of new list item (zone myzone.com) in " + rpc.path)
    new_root = data.create_node_rpc(data.get_data_root(), rpc, {"zone": {"domain": "myzone.com"}})
    new_node_ii = data.parse_ii("/dns-server:dns-server/zones/zone", PathFormat.URL)
    new_node = new_root.goto(new_node_ii)
    assert "myzone.com" in map(lambda x: x.get("domain"), new_node.value)

    rpc.path = "/dns-server:dns-server/zones/zone=myzone.com"
    rpc.path_format = PathFormat.URL

    # info("Testing creation of new leaf-list inside object " + rpc.path)
    new_root2 = data.create_node_rpc(new_root, rpc, {"access-control-list": "acl-notify-pokus"})
    new_node_ii = data.parse_ii("/dns-server:dns-server/zones/zone=myzone.com", PathFormat.URL)
    new_node2 = new_root2.goto(new_node_ii)
    assert "acl-notify-pokus" in new_node2.member("access-control-list").value


def test_nacm(datastore_1, nacm_datastore_1):
    nacm_data = nacm_datastore_1
    nacm_conf = NacmConfig(nacm_data)

    data = datastore_1
    data.register_nacm(nacm_conf)
    nacm_conf.set_ds(data)

    test_user = "dominik"

    test_paths = (
        (
            "/dns-server:dns-server/zones/zone[domain='example.com']/query-module",
            Permission.NACM_ACCESS_UPDATE,
            Action.DENY
        ),
        (
            "/dns-server:dns-server/zones/zone",
            Permission.NACM_ACCESS_READ,
            Action.PERMIT
        ),
        (
            "/dns-server:dns-server/server-options",
            Permission.NACM_ACCESS_READ,
            Action.PERMIT
        )
    )

    for test_path in test_paths:
        print("Testing path \"{}\"".format(test_path[0]))

        ii = data.parse_ii(test_path[0], PathFormat.XPATH)
        datanode = data.get_node(data.get_data_root(), ii)
        if datanode:
            print("Node found")
            # debug("Node contents: {}".format(datanode.value))
            test_ii = data.parse_ii(test_path[0], PathFormat.XPATH)
            rule = []
            action = nacm_conf.get_user_rules(test_user).check_data_node_permission(data.get_data_root(), test_ii, test_path[1],
                                                                                    out_matching_rule=rule)
            assert action == test_path[2]
            """
            if action == test_path[2]:
                info("Action = {}, OK ({})\n".format(action.name, rule[0].name if len(rule) > 0 else "default"))
            else:
                info("Action = {}, FAILED ({})\n".format(action.name, rule[0].name if len(rule) > 0 else "default"))
            """
        else:
            pytest.fail("Node not found!")

    test_ii2 = data.parse_ii("/dns-server:dns-server/zones/zone[domain='example.com']", PathFormat.XPATH)

    # info("Reading: " + str(test_ii2))
    res = nacm_conf.get_user_rules(test_user).prun_data_tree(data.get_data_root(), test_ii2)
    res = json.dumps(res.value, indent=4, sort_keys=True)

    res_expected = """
        {
        "master": [
            "server1"
        ],
        "access-control-list": [
            "acl-xfr-update",
            "acl-notify"
        ],
        "any-to-tcp": false,
        "template": "default",
        "notify": {
            "recipient": [
                "server0"
            ]
        },
        "domain": "example.com"
        }"""

    assert json.loads(res) == json.loads(res_expected)



#!/usr/bin/env python3

import sys
import os
import runpy

path = os.path.dirname(sys.modules[__name__].__file__)
path = os.path.join(path, '..')
sys.path.insert(0, path)
runpy.run_module('jetconf', run_name="__main__", alter_sys=True)




import os
import colorlog
import getopt
import logging
import sys
import signal

from importlib import import_module
from pkg_resources import resource_string, get_distribution
from colorlog import error, info
from yaml.parser import ParserError

from yangson.enumerations import ContentType, ValidationScope
from yangson.exceptions import YangsonException
from yangson.schemanode import SchemaError, SemanticError
from yangson.datamodel import DataModel

from . import op_internal
from .rest_server import RestServer
from .config import CONFIG_GLOBAL, CONFIG_NACM, load_config, validate_config, print_config
from .helpers import ErrorHelpers


def print_help():
    print("Jetconf command line options:")
    print("-c [config file]        | Pass the configuration file in YAML format")
    print("-v                      | Print version info")
    print("-h                      | Display this help")


def main():
    config_file = "config.yaml"

    # Check for Python version
    if sys.version_info < (3, 5):
        print("Jetconf requires Python version 3.5 or higher")
        sys.exit(1)

    # Get Jetconf version
    jetconf_version = get_distribution("jetconf").version

    # Parse command line arguments
    try:
        opts, args = getopt.getopt(sys.argv[1:], "c:vh")
    except getopt.GetoptError:
        print("Error: invalid argument detected.")
        print_help()
        sys.exit(1)

    for opt, arg in opts:
        if opt == "-c":
            config_file = arg
        elif opt == "-v":
            print("Jetconf version {}".format(jetconf_version))
            sys.exit(0)
        elif opt == "-h":
            print_help()
            sys.exit(0)

    # Load configuration
    try:
        load_config(config_file)
    except FileNotFoundError:
        print("Configuration file does not exist")
        sys.exit(1)
    except ParserError as e:
        print("Configuration syntax error: " + str(e))
        sys.exit(1)

    # Validate configuration
    try:
        validate_config()
    except ValueError as e:
        print("Error: " + str(e))
        sys.exit(1)

    # Set logging level
    log_level = {
        "error": logging.ERROR,
        "warning": logging.WARNING,
        "info": logging.INFO,
        "debug": logging.INFO
    }.get(CONFIG_GLOBAL["LOG_LEVEL"], logging.INFO)
    logging.root.handlers.clear()

    # Daemonize
    if CONFIG_GLOBAL["LOGFILE"] not in ("-", "stdout"):
        # Setup basic logging
        logging.basicConfig(
            format="%(asctime)s %(levelname)-8s %(message)s",
            level=log_level,
            filename=CONFIG_GLOBAL["LOGFILE"]
        )

        # Go to background
        pid = os.fork()
        if pid != 0:
            sys.exit(0)
        os.setsid()
        os.umask(0)
        pid = os.fork()
        if pid != 0:
            sys.exit(0)

        # Close standard file descriptors
        os.close(sys.stdin.fileno())
        os.close(sys.stdout.fileno())
        os.close(sys.stderr.fileno())
        fd_null = os.open("/dev/null", os.O_RDWR)
        os.dup(fd_null)
        os.dup(fd_null)
    else:
        # Setup color logging
        log_formatter = colorlog.ColoredFormatter(
            "%(asctime)s %(log_color)s%(levelname)-8s%(reset)s %(message)s",
            datefmt=None,
            reset=True,
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red',
            },
            secondary_log_colors={},
            style='%'
        )

        log_handler = colorlog.StreamHandler()
        log_handler.setFormatter(log_formatter)
        log_handler.stream = sys.stdout

        logger = colorlog.getLogger()
        logger.addHandler(log_handler)
        logger.setLevel(log_level)

    # Print version
    info("Jetconf version {}".format(jetconf_version))

    # Print configuration
    print_config()

    # Create pidfile
    fl = os.open(CONFIG_GLOBAL["PIDFILE"], os.O_WRONLY + os.O_CREAT, 0o666)
    try:
        os.lockf(fl, os.F_TLOCK, 0)
        os.write(fl, str(os.getpid()).encode())
        os.fsync(fl)
    except BlockingIOError:
        error("Jetconf daemon already running (pidfile exists). Exiting.")
        sys.exit(1)

    # Set signal handlers
    def sig_exit_handler(signum, frame):
        os.close(fl)
        os.unlink(CONFIG_GLOBAL["PIDFILE"])
        info("Exiting.")
        sys.exit(0)

    signal.signal(signal.SIGTERM, sig_exit_handler)
    signal.signal(signal.SIGINT, sig_exit_handler)

    # Import backend modules
    backend_package = CONFIG_GLOBAL["BACKEND_PACKAGE"]
    try:
        usr_state_data_handlers = import_module(backend_package + ".usr_state_data_handlers")
        usr_conf_data_handlers = import_module(backend_package + ".usr_conf_data_handlers")
        usr_op_handlers = import_module(backend_package + ".usr_op_handlers")
        usr_datastore = import_module(backend_package + ".usr_datastore")
    except ImportError as e:
        error(ErrorHelpers.epretty(e))
        error("Cannot import backend package \"{}\". Exiting.".format(backend_package))
        sys.exit(1)

    # Load data model
    yang_mod_dir = CONFIG_GLOBAL["YANG_LIB_DIR"]
    yang_lib_str = resource_string(backend_package, "yang-library-data.json").decode("utf-8")
    datamodel = DataModel(yang_lib_str, [yang_mod_dir])

    # Datastore init
    datastore = usr_datastore.UserDatastore(datamodel, CONFIG_GLOBAL["DATA_JSON_FILE"], with_nacm=CONFIG_NACM["ENABLED"])
    try:
        datastore.load()
    except (FileNotFoundError, YangsonException) as e:
        error("Cannot load JSON datastore " + CONFIG_GLOBAL["DATA_JSON_FILE"])
        error(ErrorHelpers.epretty(e))
        sig_exit_handler(0, None)

    # Validate datastore on startup
    try:
        datastore.get_data_root().validate(ValidationScope.all, ContentType.config)
    except (SchemaError, SemanticError) as e:
        error("Initial validation of datastore failed")
        error(ErrorHelpers.epretty(e))
        sig_exit_handler(0, None)

    # Register handlers for configuration data
    usr_conf_data_handlers.register_conf_handlers(datastore)

    # Register handlers for state data
    usr_state_data_handlers.register_state_handlers(datastore)

    # Register handlers for operations
    op_internal.register_op_handlers(datastore)
    usr_op_handlers.register_op_handlers(datastore)

    # Create HTTP server
    rest_srv = RestServer()
    rest_srv.register_api_handlers(datastore)
    rest_srv.register_static_handlers()

    # Run HTTP server
    rest_srv.run()


if __name__ == "__main__":
    main()



import asyncio
import ssl

from io import BytesIO
from collections import OrderedDict
from colorlog import error, warning as warn, info
from typing import List, Tuple, Dict, Callable, Optional

from h2.config import H2Configuration
from h2.connection import H2Connection
from h2.errors import ErrorCodes as H2ErrorCodes
from h2.exceptions import ProtocolError
from h2.events import DataReceived, RequestReceived, RemoteSettingsChanged, StreamEnded, WindowUpdated

from . import http_handlers as handlers
from .config import CONFIG_HTTP, API_ROOT_data, API_ROOT_RUNNING_data, API_ROOT_ops, API_ROOT_ylv
from .helpers import SSLCertT, LogHelpers
from .data import BaseDatastore
from .http_handlers import (
    HttpResponse,
    HttpStatus,
    RestconfErrType,
    ERRTAG_MALFORMED,
    ERRTAG_OPNOTSUPPORTED,
    ERRTAG_REQLARGE
)


HandlerConditionT = Callable[[str, str], bool]  # Function(method, path) -> bool
HttpHandlerT = Callable[[OrderedDict, Optional[str], SSLCertT], handlers.HttpResponse]
debug_srv = LogHelpers.create_module_dbg_logger(__name__)

h2_handlers = None  # type: HttpHandlerList


class RequestData:
    def __init__(self, headers: OrderedDict, data: BytesIO):
        self.headers = headers
        self.data = data
        self.data_overflow = False


class ResponseData:
    def __init__(self, data: bytes):
        self.data = data
        self.bytes_sent = 0


class HttpHandlerList:
    def __init__(self):
        self.handlers = []              # type: List[Tuple[HandlerConditionT, HttpHandlerT]]
        self.default_handler = None     # type: HttpHandlerT

    def register(self, condition: HandlerConditionT, handler: HttpHandlerT):
        self.handlers.append((condition, handler))

    def register_default(self, handler: HttpHandlerT):
        self.default_handler = handler

    def get_handler(self, method: str, path: str) -> HttpHandlerT:
        for h in self.handlers:
            if h[0](method, path):
                return h[1]

        return self.default_handler


class H2Protocol(asyncio.Protocol):
    def __init__(self):
        self.conn = H2Connection(H2Configuration(client_side=False, header_encoding="utf-8"))
        self.transport = None
        self.stream_data = {}       # type: Dict[int, RequestData]
        self.resp_stream_data = {}  # type: Dict[int, ResponseData]
        self.client_cert = None     # type: SSLCertT

    def connection_made(self, transport: asyncio.Transport):
        self.transport = transport
        self.client_cert = transport.get_extra_info("peercert")

        ssl_context = transport.get_extra_info("ssl_object")
        if ssl.HAS_ALPN:
            agreed_protocol = ssl_context.selected_alpn_protocol()
        else:
            agreed_protocol = ssl_context.selected_npn_protocol()

        if agreed_protocol is None:
            error("Connection error, client does not support HTTP/2")
            transport.close()
        else:
            self.conn.initiate_connection()

    def data_received(self, data: bytes):
        events = self.conn.receive_data(data)

        for event in events:
            if isinstance(event, RequestReceived):
                # Store request headers
                headers = OrderedDict(event.headers)
                request_data = RequestData(headers, BytesIO())
                self.stream_data[event.stream_id] = request_data
            elif isinstance(event, DataReceived):
                # Store incoming data
                try:
                    stream_data = self.stream_data[event.stream_id]
                except KeyError:
                    self.conn.reset_stream(event.stream_id, error_code=H2ErrorCodes.PROTOCOL_ERROR)
                else:
                    # Check if incoming data are not excessively large
                    if (stream_data.data.tell() + len(event.data)) < (CONFIG_HTTP["UPLOAD_SIZE_LIMIT"] * 1048576):
                        stream_data.data.write(event.data)
                    else:
                        stream_data.data_overflow = True
                        self.conn.reset_stream(event.stream_id, error_code=H2ErrorCodes.ENHANCE_YOUR_CALM)
            elif isinstance(event, StreamEnded):
                # Process request
                try:
                    request_data = self.stream_data.pop(event.stream_id)
                except KeyError:
                    self.send_response(
                        HttpResponse.error(HttpStatus.BadRequest, RestconfErrType.Transport, ERRTAG_MALFORMED),
                        event.stream_id
                    )
                else:
                    if request_data.data_overflow:
                        self.send_response(
                            HttpResponse.error(HttpStatus.ReqTooLarge, RestconfErrType.Transport, ERRTAG_REQLARGE),
                            event.stream_id
                        )
                    else:
                        headers = request_data.headers
                        http_method = headers[":method"]

                        if http_method in ("GET", "DELETE", "OPTIONS", "HEAD"):
                            self.run_request_handler(headers, event.stream_id, None)
                        elif http_method in ("PUT", "POST"):
                            body = request_data.data.getvalue().decode("utf-8")
                            self.run_request_handler(headers, event.stream_id, body)
                        else:
                            warn("Unknown http method \"{}\"".format(headers[":method"]))
                            self.send_response(
                                HttpResponse.error(
                                    HttpStatus.MethodNotAllowed,
                                    RestconfErrType.Transport,
                                    ERRTAG_OPNOTSUPPORTED
                                ),
                                event.stream_id
                            )
            elif isinstance(event, RemoteSettingsChanged):
                changed_settings = {}
                for s in event.changed_settings.items():
                    changed_settings[s[0]] = s[1].new_value
                self.conn.update_settings(changed_settings)
            elif isinstance(event, WindowUpdated):
                try:
                    debug_srv(
                        "str {} nw={}".format(event.stream_id, self.conn.local_flow_control_window(event.stream_id))
                    )
                    self.send_response_continue(event.stream_id)
                except (ProtocolError, KeyError) as e:
                    # debug_srv("wupdexception strid={}: {}".format(event.stream_id, str(e)))
                    pass
            # else:
            #     print(type(event))

            dts = self.conn.data_to_send()
            if dts:
                self.transport.write(dts)

    def max_chunk_size(self, stream_id: int):
        return min(self.conn.max_outbound_frame_size, self.conn.local_flow_control_window(stream_id))

    # Find and run handler for specific URI and HTTP method
    def run_request_handler(self, headers: OrderedDict, stream_id: int, data: Optional[str]):
        url_path = headers[":path"].split("?")[0].rstrip("/")
        method = headers[":method"]

        if method == "HEAD":
            h = h2_handlers.get_handler("GET", url_path)
        else:
            h = h2_handlers.get_handler(method, url_path)

        if not h:
            self.send_response(
                HttpResponse.error(HttpStatus.BadRequest, RestconfErrType.Transport, ERRTAG_MALFORMED),
                stream_id
            )
        else:
            # Run handler and send HTTP response
            resp = h(headers, data, self.client_cert)
            if method == "HEAD":
                resp.data = bytes()
            self.send_response(resp, stream_id)

    def send_response(self, resp: HttpResponse, stream_id: int):
        resp_headers = (
            (":status", resp.status_code),
            ("Content-Type", resp.content_type),
            ("Content-Length", str(resp.content_length)),
            ("Server", CONFIG_HTTP["SERVER_NAME"]),
            ("Cache-Control", "No-Cache"),
            ("Access-Control-Allow-Origin", "*"),
            ("Access-Control-Allow-Methods", "POST, GET, OPTIONS, PUT, DELETE"),
            ("Access-Control-Allow-Headers", "Content-Type")
        )

        if resp.extra_headers:
            resp_headers_od = OrderedDict(resp_headers)
            resp_headers_od.update(resp.extra_headers)
            resp_headers = resp_headers_od.items()

        self.conn.send_headers(stream_id, resp_headers)

        # Do this for optimization
        if len(resp.data) <= self.max_chunk_size(stream_id):
            self.conn.send_data(stream_id, resp.data, end_stream=True)
        else:
            self.resp_stream_data[stream_id] = ResponseData(resp.data)
            self.send_response_continue(stream_id)

    def send_response_continue(self, stream_id: int):
        resp_data = self.resp_stream_data[stream_id]
        debug_srv("Continuing...")

        while self.max_chunk_size(stream_id) != 0:
            if resp_data.bytes_sent >= len(resp_data.data):
                self.send_response_end(stream_id)
                return

            # Get available window
            chunk_size = self.max_chunk_size(stream_id)
            data_chunk = resp_data.data[resp_data.bytes_sent:resp_data.bytes_sent + chunk_size]
            resp_data.bytes_sent += chunk_size
            debug_srv("len = {}, max = {}, sent={}, dlen={}, strid={}".format(
                len(data_chunk),
                chunk_size,
                resp_data.bytes_sent,
                len(resp_data.data),
                stream_id
            ))
            self.conn.send_data(stream_id, data_chunk, end_stream=False)

    def send_response_end(self, stream_id: int):
        debug_srv("Ending stream {}...".format(stream_id))
        self.conn.send_data(stream_id, bytes(), end_stream=True)
        del self.resp_stream_data[stream_id]


class RestServer:
    def __init__(self):
        # HTTP server init
        self.http_handlers = HttpHandlerList()
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.options |= (ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1 | ssl.OP_NO_COMPRESSION)
        ssl_context.load_cert_chain(certfile=CONFIG_HTTP["SERVER_SSL_CERT"], keyfile=CONFIG_HTTP["SERVER_SSL_PRIVKEY"])
        if ssl.HAS_ALPN:
            ssl_context.set_alpn_protocols(["h2"])
        else:
            info("Python not compiled with ALPN support, using NPN instead.")
            ssl_context.set_npn_protocols(["h2"])
        if not CONFIG_HTTP["DBG_DISABLE_CERTS"]:
            ssl_context.verify_mode = ssl.CERT_REQUIRED
        ssl_context.load_verify_locations(cafile=CONFIG_HTTP["CA_CERT"])

        self.loop = asyncio.get_event_loop()

        # Each client connection will create a new H2Protocol instance
        listener = self.loop.create_server(
            H2Protocol,
            "127.0.0.1" if CONFIG_HTTP["LISTEN_LOCALHOST_ONLY"] else "",
            CONFIG_HTTP["PORT"],
            ssl=ssl_context
        )
        self.server = self.loop.run_until_complete(listener)

    def register_api_handlers(self, datastore: BaseDatastore):
        global h2_handlers

        # Register HTTP handlers
        api_get_root = handlers.api_root_handler
        api_get_ylv = handlers.api_ylv_handler
        api_get = handlers.create_get_api(datastore)
        api_get_run = handlers.create_get_running_api(datastore)
        api_post = handlers.create_post_api(datastore)
        api_put = handlers.create_put_api(datastore)
        api_delete = handlers.create_api_delete(datastore)
        api_get_op = handlers.create_api_get_op(datastore)
        api_op = handlers.create_api_op(datastore)

        self.http_handlers.register(lambda m, p: (m == "GET") and (p.startswith(API_ROOT_data)), api_get)
        self.http_handlers.register(lambda m, p: (m == "GET") and (p.startswith(API_ROOT_RUNNING_data)), api_get_run)
        self.http_handlers.register(lambda m, p: (m == "GET") and (p == API_ROOT_ylv), api_get_ylv)
        self.http_handlers.register(lambda m, p: (m == "GET") and (p == CONFIG_HTTP["API_ROOT"]), api_get_root)
        self.http_handlers.register(lambda m, p: (m == "POST") and (p.startswith(API_ROOT_data)), api_post)
        self.http_handlers.register(lambda m, p: (m == "PUT") and (p.startswith(API_ROOT_data)), api_put)
        self.http_handlers.register(lambda m, p: (m == "DELETE") and (p.startswith(API_ROOT_data)), api_delete)
        self.http_handlers.register(lambda m, p: (m == "GET") and (p.startswith(API_ROOT_ops)), api_get_op)
        self.http_handlers.register(lambda m, p: (m == "POST") and (p.startswith(API_ROOT_ops)), api_op)
        self.http_handlers.register(lambda m, p: m == "OPTIONS", handlers.options_api)

        h2_handlers = self.http_handlers

    def register_static_handlers(self):
        global h2_handlers

        self.http_handlers.register(lambda m, p: (m == "GET") and not (p.startswith(CONFIG_HTTP["API_ROOT"])), handlers.get_file)
        self.http_handlers.register_default(handlers.unknown_req_handler)

        h2_handlers = self.http_handlers

    def run(self):
        info("Server started on {}".format(self.server.sockets[0].getsockname()))
        self.loop.run_forever()

    def shutdown(self):
        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())
        self.loop.close()



# This script reads zone data from KnotDNS socket and converts them
# to YANG model compliant data tree in JSON formatting.
# Only SOA, A, AAAA, NS, MX, TXT, TLSA and CNAME records are currently
# supported.

import sys
import json

from typing import Dict, Any
from libknot.control import KnotCtl, KnotCtlType

# Edit this to match your actual KnotDNS control socket
KNOT_SOCKET = "/home/pspirek/knot-conf/knot.sock"


def main(args):
    if len(args) != 2:
        print("Usage: {} [domain]".format(args[0]))
        exit(1)

    domain = args[1]    # type: str
    if domain[-1] != '.':
        domain += "."

    ctl = KnotCtl()
    ctl.connect(KNOT_SOCKET)
    ctl.send_block("zone-read", zone=domain)
    resp = ctl.receive_block()  # type: Dict[str, Any]
    ctl.send(KnotCtlType.END)
    ctl.close()

    resp = resp[domain]

    zone_template = {
        "dns-zones:zone-data": {
            "zone": [
                {
                    "name": domain,
                    "class": "IN",
                    "default-ttl": 3600,
                    "SOA": {},
                    "rrset": []
                }
            ]
        }
    }

    zone_out = zone_template["dns-zones:zone-data"]["zone"][0]
    soa_out = zone_out["SOA"]

    soa = resp[domain]["SOA"]
    soa_data = soa["data"][0].split()

    try:
        soa_out["ttl"] = int(soa["ttl"])
        soa_out["mname"] = soa_data[0]
        soa_out["rname"] = soa_data[1]
        soa_out["serial"] = int(soa_data[2])
        soa_out["refresh"] = int(soa_data[3])
        soa_out["retry"] = int(soa_data[4])
        soa_out["expire"] = int(soa_data[5])
        soa_out["minimum"] = int(soa_data[6])
    except (IndexError, ValueError) as e:
        print(str(e))

    rrset_out = zone_out["rrset"]

    for owner, rrs in resp.items():
        # print("rrs={}".format(rrs))
        for rr_type, rr in rrs.items():
            # print("rr={}".format(rr))
            if rr_type not in ("A", "AAAA", "NS", "MX", "TXT", "TLSA", "CNAME"):
                continue

            ttl = int(rr["ttl"])
            rr_data_list = rr["data"]

            new_rr_out_rdata_list = []
            new_rr_out = {
                "owner": owner,
                "type": "iana-dns-parameters:" + rr_type,
                "ttl": ttl,
                "rdata": new_rr_out_rdata_list
            }

            id_int = 0
            for rr_data in rr_data_list:
                new_rr_out_rdata_values = {}
                new_rr_out_rdata = {
                    "id": str(id_int),
                    rr_type: new_rr_out_rdata_values
                }

                if rr_type in ("A", "AAAA"):
                    new_rr_out_rdata_values["address"] = rr_data
                elif rr_type == "NS":
                    new_rr_out_rdata_values["nsdname"] = rr_data
                elif rr_type == "MX":
                    rr_data = rr_data.split()
                    new_rr_out_rdata_values["preference"] = rr_data[0]
                    new_rr_out_rdata_values["exchange"] = rr_data[1]
                elif rr_type == "TXT":
                    new_rr_out_rdata_values["txt-data"] = rr_data.strip(" \"")
                elif rr_type == "TLSA":
                    cert_usage_enum = {
                        "0": "PKIX-TA",
                        "1": "PKIX-EE",
                        "2": "DANE-TA",
                        "3": "DANE-EE",
                        "255": "PrivCert"
                    }
                    sel_enum = {
                        "0": "Cert",
                        "1": "SPKI",
                        "255": "PrivSel"
                    }
                    match_type_enum = {
                        "0": "Full",
                        "1": "SHA2-256",
                        "2": "SHA2-512",
                        "255": "PrivMatch"
                    }
                    rr_data = rr_data.split()
                    new_rr_out_rdata_values["certificate-usage"] = cert_usage_enum[rr_data[0]]
                    new_rr_out_rdata_values["selector"] = sel_enum[rr_data[1]]
                    new_rr_out_rdata_values["matching-type"] = match_type_enum[rr_data[2]]
                    new_rr_out_rdata_values["certificate-association-data"] = rr_data[3]
                elif rr_type == "CNAME":
                    new_rr_out_rdata_values["cname"] = rr_data

                new_rr_out_rdata_list.append(new_rr_out_rdata)
                id_int += 1

            rrset_out.append(new_rr_out)

    print(json.dumps(zone_template, indent=4, sort_keys=True))

if __name__ == "__main__":
    main(sys.argv)



from typing import List, Tuple, Callable, Any

from yangson.schemanode import SchemaNode
from yangson.schemadata import SchemaData
from yangson.instance import InstanceRoute

from .helpers import JsonNodeT


HandlerSelectorT = Any


# ---------- Conf data base objects ----------

class ConfDataObjectHandler:
    def __init__(self, ds: "BaseDatastore", sch_pth: str):
        self.ds = ds
        self.schema_path = sch_pth                          # type: str
        self.schema_node = ds.get_schema_node(sch_pth)      # type: SchemaNode

    def create(self, ii: InstanceRoute, ch: "DataChange"):
        pass

    def replace(self, ii: InstanceRoute, ch: "DataChange"):
        pass

    def delete(self, ii: InstanceRoute, ch: "DataChange"):
        pass

    def __str__(self):
        return self.__class__.__name__ + ": listening at " + self.schema_path


class ConfDataListHandler:
    def __init__(self, ds: "BaseDatastore", sch_pth: str):
        self.ds = ds
        self.schema_path = sch_pth                          # type: str
        self.schema_node = ds.get_schema_node(sch_pth)      # type: SchemaNode

    def create_item(self, ii: InstanceRoute, ch: "DataChange"):
        pass

    def replace_item(self, ii: InstanceRoute, ch: "DataChange"):
        pass

    def delete_item(self, ii: InstanceRoute, ch: "DataChange"):
        pass

    def create_list(self, ii: InstanceRoute, ch: "DataChange"):
        pass

    def replace_list(self, ii: InstanceRoute, ch: "DataChange"):
        pass

    def delete_list(self, ii: InstanceRoute, ch: "DataChange"):
        pass

    def __str__(self):
        return self.__class__.__name__ + ": listening at " + self.schema_path


# ---------- State data base objects ----------

class StateDataHandlerBase:
    def __init__(self, datastore: "BaseDatastore", schema_path: str):
        self.ds = datastore
        self.data_model = datastore.get_dm()
        self.sch_pth = schema_path
        self.schema_node = self.data_model.get_data_node(self.sch_pth)


class StateDataContainerHandler(StateDataHandlerBase):
    def generate_node(self, node_ii: InstanceRoute, username: str, staging: bool) -> JsonNodeT:
        pass


class StateDataListHandler(StateDataHandlerBase):
    def generate_list(self, node_ii: InstanceRoute, username: str, staging: bool) -> JsonNodeT:
        pass

    def generate_item(self, node_ii: InstanceRoute, username: str, staging: bool) -> JsonNodeT:
        pass


# ---------- Handles lists ----------

class BaseHandlerList:
    def __init__(self):
        self.handlers = []              # type: List[Tuple[HandlerSelectorT, Callable]]
        self.default_handler = None     # type: Callable

    def register(self, identifier: str, handler: Callable):
        raise NotImplementedError("Not implemented in base class")

    def register_default(self, handler: Callable):
        self.default_handler = handler

    def get_handler(self, identifier: str) -> Any:
        raise NotImplementedError("Not implemented in base class")


class OpHandlerList(BaseHandlerList):
    def register(self, handler: Callable, op_name: str):
        self.handlers.append((op_name, handler))

    def get_handler(self, op_name: str) -> Callable:
        for h in self.handlers:
            if h[0] == op_name:
                return h[1]

        return self.default_handler


class ConfDataHandlerList:
    def __init__(self):
        self.handlers = []  # type: List[Tuple[HandlerSelectorT, BaseDataListener]]

    def register(self, handler: "BaseDataListener"):
        schema_node = handler.schema_node  # type: SchemaNode
        sch_node_id = str(id(schema_node))
        self.handlers.append((sch_node_id, handler))

    def get_handler(self, sch_node_id: str) -> "BaseDataListener":
        for h in self.handlers:
            if h[0] == sch_node_id:
                return h[1]

        return None


class StateDataHandlerList:
    def __init__(self):
        self.handlers = []

    def register(self, handler: "StateNodeHandlerBase"):
        saddr = SchemaData.path2route(handler.sch_pth)
        self.handlers.append((saddr, handler))

    def get_handler(self, sch_pth: str, allow_superior: bool = True) -> Callable:
        saddr = SchemaData.path2route(sch_pth)
        if allow_superior:
            while saddr:
                for h in self.handlers:
                    if h[0] == saddr:
                        return h[1]
                saddr.pop()
        else:
            for h in self.handlers:
                if h[0] == saddr:
                    return h[1]

        return None


# ---------- Handler list globals ----------

OP_HANDLERS = OpHandlerList()
STATE_DATA_HANDLES = StateDataHandlerList()
CONF_DATA_HANDLES = ConfDataHandlerList()



from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(here, 'README.rst'), encoding='utf-8') as readme:
    long_description = readme.read()

setup(
    name = "jetconf",
    packages = find_packages(),
    use_scm_version = True,
    setup_requires=["setuptools_scm"],
    description = "Pure Python implementation of RESTCONF server",
    long_description = long_description,
    url = "https://gitlab.labs.nic.cz/labs/jetconf",
    author = "Pavel Spirek",
    author_email = "pavel.spirek@nic.cz",
    entry_points = {
        "console_scripts": ["jetconf=jetconf.__main__:main"]
    },
    install_requires = ["yangson", "h2", "colorlog", "pyaml", "pytz"],
    tests_require = ["pytest"],
    keywords = ["RESTCONF", "yang", "data model", "configuration", "json"],
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: System :: Monitoring",
        "Topic :: System :: Systems Administration"
    ]
)



from .config import CONFIG
from .helpers import JsonNodeT
from .handler_list import OP_HANDLERS
from .data import BaseDatastore, RpcInfo, StagingDataException


class OpHandlersContainer:
    def __init__(self, ds: BaseDatastore):
        self.ds = ds

    def jetconf_conf_start(self, rpc: RpcInfo) -> JsonNodeT:
        # try:
        #     transaction_opts = rpc.op_input_args["options"]
        # except (TypeError, KeyError):
        #     transaction_opts = None
        # self.ds.make_user_journal(rpc.username, transaction_opts)
        self.ds.make_user_journal(rpc.username, None)
        ret_data = {"status": "OK"}

        return ret_data

    def jetconf_conf_status(self, rpc: RpcInfo) -> JsonNodeT:
        try:
            usr_journal = self.ds.get_user_journal(rpc.username)
            transaction_opened = True
        except StagingDataException:
            transaction_opened = False

        ret_data = {
            "status": "OK",
            "transaction-opened": transaction_opened
        }

        return ret_data

    def jetconf_conf_reset(self, rpc: RpcInfo) -> JsonNodeT:
        self.ds.drop_user_journal(rpc.username)
        ret_data = {"status": "OK"}
        return ret_data

    def jetconf_conf_commit(self, rpc: RpcInfo) -> JsonNodeT:
        try:
            usr_journal = self.ds.get_user_journal(rpc.username)
        except StagingDataException:
            usr_journal = None

        if usr_journal is not None:
            try:
                self.ds.lock_data(rpc.username)
                commit_res = usr_journal.commit(self.ds)
                if CONFIG["GLOBAL"]["PERSISTENT_CHANGES"] is True:
                    self.ds.save()
            finally:
                self.ds.unlock_data()

            self.ds.drop_user_journal(rpc.username)
        else:
            commit_res = False

        ret_data = {
            "status": "OK",
            "conf-changed": commit_res
        }

        return ret_data

    def jetconf_get_schema_digest(self, rpc: RpcInfo) -> JsonNodeT:
        ret_data = self.ds.get_dm().schema_digest()
        return ret_data

    def jetconf_get_list_length(self, rpc: RpcInfo) -> JsonNodeT:
        try:
            list_url = rpc.op_input_args["url"]  # type: str
        except (TypeError, KeyError):
            raise ValueError("This operation expects \"url\" input parameter")

        try:
            staging = rpc.op_input_args["staging"]  # type: str
        except (TypeError, KeyError):
            staging = False

        rpc_gll = RpcInfo()
        rpc_gll.username = rpc.username
        rpc_gll.skip_nacm_check = rpc.skip_nacm_check
        rpc_gll.path = list_url.rstrip("/")
        rpc_gll.qs = {}

        ln_val = self.ds.get_node_rpc(rpc_gll, staging).value

        if isinstance(ln_val, list):
            ret_data = {"jetconf:list-length": len(ln_val)}
        else:
            raise ValueError("Passed URI does not point to List")

        return ret_data


def register_op_handlers(ds: BaseDatastore):
    op_handlers_obj = OpHandlersContainer(ds)
    # OP_HANDLERS.register(op_handlers_obj.jetconf_conf_start, "jetconf:conf-start")
    OP_HANDLERS.register(op_handlers_obj.jetconf_conf_status, "jetconf:conf-status")
    OP_HANDLERS.register(op_handlers_obj.jetconf_conf_reset, "jetconf:conf-reset")
    OP_HANDLERS.register(op_handlers_obj.jetconf_conf_commit, "jetconf:conf-commit")
    OP_HANDLERS.register(op_handlers_obj.jetconf_get_schema_digest, "jetconf:get-schema-digest")
    OP_HANDLERS.register(op_handlers_obj.jetconf_get_list_length, "jetconf:get-list-length")
